import os
import glob
import logging
try:
    import h5py
    pass
except:
    logging.info("import h5py failed")
import numpy as np
import sys

def get_data(datafile):
    #get data for training
    f=h5py.File(datafile,'r')
    y=f.get('target')
    X=np.array(f.get('ECAL'))
    y=(np.array(y[:,1]))
    X[X < 1e-4] = 0
    X = np.expand_dims(X, axis=-1)
    X = X.astype(np.float32)
    y = y.astype(np.float32)
    y = y/100.
    ecal = np.squeeze(np.sum(X, axis=(1, 2, 3)))
    logging.info("X shape {}; y shape {}; ecal shape {}".format(str(X.shape)), str(y.shape), str(ecal.shape))

    f.close()
    return X, y, ecal

dest='/data/shared/3DGAN/'
import socket
host = os.environ.get('HOST', os.environ.get('HOSTNAME',socket.gethostname()))
if 'daint' in host:
    dest='/scratch/snx3000/vlimant/3DGAN/'
if 'titan' in host:
    dest='/ccs/proj/csc291/DATA/3DGAN/'

sub_split = int(sys.argv[1]) if len(sys.argv)>1 else 1
                                          
for F in glob.glob('/bigdata/shared/LCD/NewV1/*scan/*.h5'):
    _,d,f = F.rsplit('/',2)
    if not 'Ele' in d: continue
    X = None
    if sub_split==1:
        nf = '%s/%s_%s.h5'%( dest,d,f)
        if os.path.isfile( nf) :
            continue
        logging.info("processing files {} into {}".format(F,nf))
        if X is None:
            X,y,ecal = get_data(F)
        o = h5py.File(nf,'w')
        o['X'] = X
        o.create_group("y")
        o['y']['a'] = np.ones(y.shape)
        o['y']['b'] = y
        o['y']['c'] = ecal
        o.close()        
    else:
        for sub in range(sub_split):
            nf = '%s/%s_%s_sub%s.h5'%(dest, d,f,sub)
            if os.path.isfile( nf) :
                continue
            logging.info("processing files {} into {}".format(F,nf))
            if X is None:
                X,y,ecal = get_data(F)
                N = X.shape[0]
                splits = [i*N/sub_split for i in range(sub_split)]+[-1]
            o = h5py.File(nf,'w')
            o['X'] = X[splits[sub]:splits[sub+1],...]
            o.create_group("y")
            o['y']['a'] = np.ones(y[splits[sub]:splits[sub+1],...].shape)
            o['y']['b'] = y[splits[sub]:splits[sub+1],...]
            o['y']['c'] = ecal[splits[sub]:splits[sub+1],...]
            o.close()
            X = None

if sub_split == 1:
    sub_files = lambda f:not 'sub' in f
else:
    sub_files = lambda f:'sub' in f

open('train_3d.list','w').write( '\n'.join(filter(sub_files,glob.glob(dest+'/*.h5')[:-4])))
open('test_3d.list','w').write( '\n'.join(filter(sub_files,glob.glob(dest+'/*.h5')[-4:])))

open('train_small_3d.list','w').write( '\n'.join(filter(sub_files,glob.glob(dest+'/*.h5')[:-4])))
open('test_small_3d.list','w').write( '\n'.join(filter(sub_files,glob.glob(dest+'/*.h5')[-4:])))

open('train_7_3d.list','w').write( '\n'.join(filter(sub_files,glob.glob(dest+'/*.h5')[:7])))
open('test_1_3d.list','w').write( '\n'.join(filter(sub_files,glob.glob(dest+'/*.h5')[-1:])))
    



### Builds one of the available models.  
# Saves model architecture to <model_name>_arch.json
# and model weights to <model_name>_weights.h5

import os
os.environ['CUDA_VISIBLE_DEVICES']=""
import argparse
import logging

from Models import make_model

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('model_name', help='model to construct')
    parser.add_argument('model_args', nargs='*', help='key=value to pass to the model',default=[])
    args = parser.parse_args()
    model_name = args.model_name
    model_args = {}
    for kw in args.model_args:
        k,v = kw.split('=')
        try:
            v = int(v)
        except:
            v= float(v)
        model_args[k] = v
    if model_args:
        logging.info("passing {} to the model builder".format(str(model_args)))
        model = make_model( model_name ,**model_args)
    else:
        model = make_model( model_name)
    weights_filename = "%s_weights.h5" % model_name
    arch_filename = "%s_arch.json" % model_name

    if not "torch" in model_name:
        model.summary()
        model.save_weights( weights_filename, overwrite=True )
        logging.info("Saved model weights to {0}".format(weights_filename))

        model_arch = model.to_json()
        with open( arch_filename, 'w' ) as arch_file:
            arch_file.write( model_arch )
        logging.info("Saved model architecture to {0}".format(arch_filename))
    else:
        import torch
        weights_filename = weights_filename.replace('h5','torch')
        arch_filename = arch_filename.replace('json','torch')
        torch.save(model.state_dict(), weights_filename)
        logging.info("Saved model weights to {0}".format(weights_filename))
        torch.save(model, arch_filename)
        logging.info("Saved model architecture to {0}".format(arch_filename))
                        



import os,sys,json
import numpy as np
import socket
import time
import logging

from .process import MPIWorker, MPIMaster
from ..util.timeline import Timeline, timeline

class MPISingleWorker(MPIWorker):
    """This class trains its model with no communication to other processes"""
    def __init__(self, num_epochs, data, algo, model_builder,
                verbose, monitor, custom_objects,
                early_stopping, target_metric,
                checkpoint, checkpoint_interval):

        self.has_parent = False

        self.best_val_loss = None
        self.target_metric = (target_metric if type(target_metric)==tuple else tuple(map(lambda s : float(s) if s.replace('.','').isdigit() else s, target_metric.split(',')))) if target_metric else None
        self.patience = (early_stopping if type(early_stopping)==tuple else tuple(map(lambda s : float(s) if s.replace('.','').isdigit() else s, early_stopping.split(',')))) if early_stopping else None
        
        logging.info("Creating MPISingleWorker")
            
        super(MPISingleWorker, self).__init__(data, algo, model_builder, process_comm=None, parent_comm=None, parent_rank=None, 
            num_epochs=num_epochs, verbose=verbose, monitor=monitor, custom_objects=custom_objects,
            checkpoint=checkpoint, checkpoint_interval=checkpoint_interval)

    def train(self):
        Timeline.begin("train")
        self.start_time = time.time()
        self.check_sanity()

        for epoch in range(1, self.num_epochs + 1):
            logging.info("beginning epoch {:d}".format(self.epoch + epoch))
            Timeline.begin("epoch")
            if self.monitor:
                self.monitor.start_monitor()
            epoch_metrics = np.zeros((1,))
            i_batch = 0

            for i_batch, batch in enumerate(self.data.generate_data()):
                Timeline.begin("train_on_batch")
                train_metrics = self.model.train_on_batch( x=batch[0], y=batch[1] )
                Timeline.end("train_on_batch")
                if epoch_metrics.shape != train_metrics.shape:
                    epoch_metrics = np.zeros( train_metrics.shape)
                epoch_metrics += train_metrics

                ######
                self.update = self.algo.compute_update( self.weights, self.model.get_weights() )
                if self.algo.mode == 'gem':
                    self.update = self.algo.compute_update_worker(self.weights, self.update)

                self.weights = self.algo.apply_update( self.weights, self.update )
                self.algo.set_worker_model_weights( self.model, self.weights )
                ######

                if self._short_batches and i_batch>self._short_batches: break

            if self.monitor:
                self.monitor.stop_monitor()
            epoch_metrics = epoch_metrics / float(i_batch+1)
            l = self.model.get_logs( epoch_metrics )
            self.update_history( l )

            Timeline.begin("epoch")

            if self.stop_training:
                break

            self.validate()
            self.save_checkpoint()
            

        logging.info("Signing off")
        if self.monitor:
            self.update_monitor( self.monitor.get_stats() )        

        self.stop_time = time.time()
        Timeline.end("train")
        self.data.finalize()

    def validate(self):
        return MPIMaster.validate_aux(self, self.weights, self.model)

    def record_details(self, json_name=None, meta=None):
        MPIMaster.record_details(self, json_name, meta)



import time 
import numpy as np
import os
import hashlib
import logging
import glob
from ..mpi.manager import MPIKFoldManager
from ..util.logger import set_logging_prefix
from ..util.utils import opt_tag_lookup

class ProcessBlock(object):
    """
    This class represents a block of processes that run model training together.

    Attributes:
    comm_world: MPI communicator with all processes.
        Used to communicate with process 0, the coordinator
    comm_block: MPI communicator with the processes in this block.
        Rank 0 is the master, other ranks are workers.
    algo: MPI Algo object
    data: MPI Data object
    device: string indicating which device (cpu or gpu) should be used
    epochs: number of training epochs
    train_list: list of training data files
    val_list: list of validation data files
    verbose: print detailed output from underlying training machinery
    """

    def __init__(self, comm_world, comm_block, algo, data, device, model_provider,
                 epochs, train_list, val_list, folds=1,
                 num_masters=1,
                 num_process=1,
                 verbose=False,
                 early_stopping=None,
                 target_metric=None,
                 monitor=False,
                 label = None,
                 restore = False,
                 checkpoint=None,
                 checkpoint_interval=5):
        set_logging_prefix(
            comm_world.Get_rank(),
            comm_block.Get_rank() if comm_block is not None else '-',
            '-',
            'B'
        )
        logging.debug("Initializing ProcessBlock")
        self.comm_world = comm_world
        self.comm_block = comm_block
        self.folds = folds
        self.num_masters = num_masters
        self.num_process = num_process
        self.algo = algo
        self.data = data
        self.device = device
        self.model_provider = model_provider
        self.epochs = epochs
        self.train_list = train_list
        self.val_list = val_list
        self.verbose = verbose
        self.last_params = None
        self.early_stopping=early_stopping
        self.target_metric=target_metric
        self.monitor = monitor
        self.current_builder = None
        self.restore = restore
        self.checkpoint = checkpoint
        self.checkpoint_interval = checkpoint_interval
        self.label = checkpoint if checkpoint else label
        
    def wait_for_model(self):
        """
        Blocks until the parent sends a parameter set
        indicating the model that should be trained.
        """
        logging.debug("Waiting for model params")
        self.last_params = self.comm_world.recv(source=0, tag=opt_tag_lookup('params'))
        params = self.last_params
        if params is not None:
            logging.debug("Received parameters {}".format(params))
            model_builder = self.model_provider.builder(*params)
            if model_builder:
                model_builder.comm = self.comm_block
                self.current_builder = model_builder
            else:
                self.current_builder = None
            return True
        return False

    def train_model(self):
        if self.current_builder is None:
            # Invalid model, return nonsense FoM
            return np.nan
        fake_train = False
        if fake_train:
            if self.comm_block.Get_rank() == 0:
                    time.sleep(abs(np.random.randn()*30))
                    result = np.random.randn()
                    logging.debug("Finished training with result {}".format(result))
                    return result
        else:
            logging.debug("Creating a manager")
            history_name = '{}-block-{}'.format(self.label,
                                                hashlib.md5(str(self.last_params).encode('utf-8')).hexdigest())
            ## need to reset this part to avoid cached values
            self.algo.reset()
            if self.restore:
                if os.path.isfile(history_name + '.latest'):
                    with open(history_name + '.latest', 'r') as latest:
                        restore_name = latest.read().splitlines()[-1]
                else:
                    restore_name = history_name
                if any([os.path.isfile(ff) for ff in glob.glob('./*' + restore_name + '.model')]):
                    self.current_builder.weights = restore_name + '.model'
                self.algo.load(restore_name)
                self.restore= False
            manager = MPIKFoldManager( self.folds,
                                       self.comm_block, self.data, self.algo, self.current_builder,
                                       self.epochs, self.train_list, self.val_list,
                                       num_masters=self.num_masters,
                                       num_process=self.num_process,
                                       verbose=self.verbose,
                                       early_stopping=self.early_stopping,
                                       target_metric=self.target_metric,
                                       monitor=self.monitor,
                                       checkpoint=history_name if self.checkpoint else None,
                                       checkpoint_interval=self.checkpoint_interval)
            manager.train()
            fom = manager.figure_of_merit()
            manager.manager.process.record_details(
                json_name=history_name + '.json',
                meta={'parameters': list(map(float,self.last_params)),
                                                         'fold' : manager.fold_num})
            manager.close()
            return fom

    def send_result(self, result):
        if self.comm_block.Get_rank() == 0:
            ## only the rank=0 in the block is sending back his fom
            logging.debug("Sending result {} to coordinator".format(result))
            self.comm_world.isend(result, dest=0, tag=opt_tag_lookup('result')) 

    def run(self):
        """
        Awaits instructions from the parent to train a model.
        Then trains it and returns the loss to the parent.
        """
        while True:
            self.comm_block.Barrier()
            logging.debug("Waiting for model")
            have_builder = self.wait_for_model()
            if not have_builder:
                logging.debug("Received exit signal from coordinator")
                break
            
            logging.debug("Will train model")
            fom = self.train_model()
            logging.debug("Done training, will send result if needed")
            self.send_result(fom)
        self.comm_world.Barrier()





import skopt 
import random
import json
import os
import pickle
import time
import hashlib
import numpy as np
from mpi4py import MPI
from .genetic_algorithm import GA
from ..util.utils import opt_tag_lookup
from ..util.logger import set_logging_prefix
import logging

class Coordinator(object):
    """
    This class coordinates the Bayesian optimization procedure.
    
    Attributes:
    comm: MPI communicator object containing all processes.
    num_blocks: int, number of blocks of workers
    opt_params: list of parameter ranges, suitable to pass
        to skopt.Optimizer's constructor
    optimizer: skopt.Optimizer instance
    model_fn: python function.  should take a list of parameters
        and return a JSON string representing the keras model to train
    param_list: list of parameter sets tried
    fom_list: list of figure of merit (FOM) values obtained
    block_dict: dict keeping track of which block is running which model
    req_dict: dict holding MPI receive requests for each running block
    best_params: best parameter set found by Bayesian optimization
    """
    
    def __init__(self, comm, num_blocks,
                 opt_params, ga, populationSize, checkpointing, label):
        set_logging_prefix(MPI.COMM_WORLD.Get_rank(), process_type='C')
        logging.debug("Coordinator initializing")
        self.comm = comm
        self.num_blocks = num_blocks

        self.opt_params = opt_params
        self.ga = ga
        if ga:
            self.populationSize = populationSize
            self.optimizer = GA(self.opt_params, self.populationSize)
        else:
            self.optimizer = skopt.Optimizer(dimensions=self.opt_params, random_state= 13579)
        self.param_list = []
        self.fom_list = []
        self.block_dict = {}
        self.req_dict = {}
        self.best_params = None
        self.best_fom = None

        self.next_params = []
        self.to_tell = []
        self.ends_cycle = False
        self.target_fom = None
        self.history = {}
        self.checkpointing = checkpointing
        self.label = checkpointing if checkpointing else label
        self.iter = 1

    def ask(self, n_iter):
        if not self.next_params:
            ## don't ask every single time
            if self.ga:
                self.next_params = self.optimizer.ask()
            else:
                self.next_params = self.optimizer.ask( n_iter )
        return self.next_params.pop(-1)

    def record_details(self, json_name=None):
        if json_name is None:
            json_name = '{}-coordinator.json'.format(self.label)
        with open(json_name, 'w') as out:
            out.write( json.dumps( self.history, indent=2))
            
    def save(self, fn = None):
        if fn is None:
            fn = '{}-coordinator.state'.format(self.label)
        self.history.setdefault('save', fn)
        with open(fn, 'wb') as state:
            self_dict = dict([(k,v) for (k,v) in self.__dict__.items() if k not in ['comm','req_dict']])
            logging.debug("Saving on-going parameters {}".format(self_dict.get('block_dict').values()))
            pickle.dump( self_dict, state )

    def load(self, fn=None):
        if fn is None:
            fn = '{}-coordinator.state'.format(self.label)
        if os.path.isfile(fn):
            self.history.setdefault('load',fn)
            with open(fn, 'rb') as state:
                logging.info("Loading the coordinator from {}".format(fn))
                self_dict = pickle.load(state)
                active_params = list(self_dict.pop('block_dict').values())
                logging.debug("Loading active params {}".format(active_params))
                self.__dict__.update(self_dict)
                self.next_params.extend(active_params)
        else:
            logging.warning('Failed to load coordinator state from {}, starting from scratch'.format(fn))

    def fit(self, step):
        X = [o[0] for o in self.to_tell]
        Y = [o[1] for o in self.to_tell]
        if X and Y:
            if self.ga:
                opt_result = self.optimizer.tell( X, Y, step//self.populationSize )
                self.best_params = opt_result[0]
                self.best_fom = opt_result[1]
            else:
                logging.debug("Fitting from {} values".format( len (X)))
                opt_result = self.optimizer.tell( X, Y )
                self.best_params = opt_result.x
                self.best_fom = opt_result.fun
                logging.info("New best param estimate, with telling {} points : {}, with value {}".format(len(X),self.best_params, self.best_fom))
                self.next_params = []
            self.to_tell = []
            ## checkpoint your self
            if self.checkpointing:
                logging.info("Checkpointing the coordinator")                
                self.save()
            self.history.setdefault('tell',[]).append({'X': [list(map(float,x)) for x in X], 'Y':Y,
                                                       'Hash' : [hashlib.md5(str(x).encode('utf-8')).hexdigest() for x in X],
                                                       'fX': list(map(float,self.best_params)), 'fY': self.best_fom})
            if self.target_fom and opt_result.fun < self.target_fom:
                logging.info("The optimization has reached the desired value at optimum {}".format(self.target_fom))
                self.ends_cycle = True

    def tell(self, params, result, step):
        self.to_tell.append( (params, result) )
        tell_right_away = False
        if tell_right_away:
            self.fit(step)
        
    def run(self, num_iterations=1):
        loopMax = num_iterations
        if self.ga:
            self.optimizer.setGenerations(num_iterations)
            loopMax *= self.populationSize
        step = self.iter
        for step in range(self.iter, loopMax + 1):
            next_block = self.wait_for_idle_block(step)
            logging.info("Coordinator iteration {}".format(step))
            if self.ends_cycle:
                logging.info("Coordinator is skiping the iteration cycle")
                break
            next_params = self.ask( num_iterations )
            logging.info("Next block: {}, next params {}".format(next_block, next_params))
            self.run_block(next_block, next_params, step)
            if self.checkpointing:
                logging.info("Checkpointing the coordinator after run_block")
                self.save()

        ## wait for all running block to finish their processing
        self.close_blocks(step)
        self.fit(step)
        
        ## end all processes
        for proc in range(1, self.comm.Get_size()):
            logging.debug("Signaling process {} to exit".format(proc))
            self.comm.send(None, dest=proc, tag=opt_tag_lookup('params'))
        self.comm.Barrier()
        logging.info("Finished all iterations!")
        logging.info("Best parameters found: {} with value {}".format(self.best_params, self.best_fom))
        
    def wait_for_idle_block(self, step):
        """
        In a loop, checks each block of processes to see if it's
        idle.  This function blocks until there is an available process.
        """
        blocklist = list(range(1, self.num_blocks+1))
        
        while True:
            self.fit(step)
            random.shuffle( blocklist ) ## look at them in random order
            for cur_block in blocklist:
                idle = self.check_block(cur_block, step)
                if idle:
                    logging.debug("From coordinator, block {} is found idling, and can be used next".format( cur_block))
                    return cur_block

    def close_blocks(self, step):
        while self.block_dict:
            logging.debug("Closing blocks, {} blocks still running".format(len(self.block_dict)))
            for block_num in list(self.block_dict.keys()):
                self.check_block( block_num, step)
            time.sleep(5)
        
    def check_block(self, block_num, step):
        """
        If the indicated block has completed a training run, store the results.
        Returns True if the block is ready to train a new model, False otherwise.
        """
        if block_num in self.block_dict:
            done, result = self.req_dict[block_num].test()
            if done:
                params = self.block_dict.pop(block_num)
                if np.isnan(result):
                    logging.info("Skipped telling due to invalid model params {}".format(params))
                else:
                    self.param_list.append(params)
                    self.fom_list.append(result)
                    logging.debug("Telling {} at {}".format(result, params))
                    self.tell( params, result, step )
                del self.req_dict[block_num]
                self.iter += 1
                return True
            return False
        else:
            return True

    def run_block(self, block_num, params, step):
        self.block_dict[block_num] = params
        # In the current setup, we need to signal each GPU in the 
        # block to start training
        block_size = int((self.comm.Get_size()-1)/self.num_blocks)
        start = (block_num-1) * block_size + 1 
        end = block_num * block_size 
        logging.debug("Launching block {}. Sending params to nodes from {} to {}".format(block_num, start,end))
        for proc in range(start, end+1):
            self.comm.send(params, dest=proc, tag=opt_tag_lookup('params')) 
        self.req_dict[block_num] = self.comm.irecv(source=start, tag=opt_tag_lookup('result'))



from mpi4py import MPI
import time
import logging
from os.path import abspath
from functools import wraps
import inspect
import importlib
import numpy as np
import re

TRACE = logging.DEBUG - 5

level_map = {
    'trace': TRACE,
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warn': logging.WARNING,
    'error': logging.ERROR,
}

base_prefix = "{ptype} {world}:{parent}:{process}"
file_handler = None
stream_handler = None

start_time = time.time()

def addLoggingLevel(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.
    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
       raise AttributeError('{} already defined in logging module'.format(levelName))
    if hasattr(logging, methodName):
       raise AttributeError('{} already defined in logging module'.format(methodName))
    if hasattr(logging.getLoggerClass(), methodName):
       raise AttributeError('{} already defined in logger class'.format(methodName))

    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)
    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)

class ElapsedTimeFormatter(logging.Formatter):
    def formatTime(self, record, datefmt=None):
        global start_time
        total_millis = int(record.created - start_time) * 1000 + int(record.msecs)

        millis = total_millis%1000
        millis = int(millis)
        seconds=(total_millis/1000)%60
        seconds = int(seconds)
        minutes=(total_millis/(1000*60))%60
        minutes = int(minutes)
        hours=(total_millis/(1000*60*60))%24
        hours = int(hours)

        elapsed = "{:04d}:{:02d}:{:02d}.{:03d}".format(hours, minutes, seconds, millis)
        return elapsed

class MPIFileHandler(logging.FileHandler):
    def __init__(self,
                 filename,
                 mode=MPI.MODE_WRONLY|MPI.MODE_CREATE|MPI.MODE_APPEND ,
                 encoding='utf-8',
                 delay=False,
                 comm=MPI.COMM_WORLD):
        self.baseFilename = abspath(filename)
        self.mode = mode
        self.encoding = encoding
        self.comm = comm.Dup()
        if delay:
            #We don't open the stream, but we still need to call the
            #Handler constructor to set level, formatter, lock etc.
            logging.Handler.__init__(self)
            self.stream = None
        else:
           logging.StreamHandler.__init__(self, self._open())

    def _open(self):
        stream = MPI.File.Open( self.comm, self.baseFilename, self.mode )
        stream.Set_atomicity(True)
        return stream

    def emit(self, record):
        try:
            msg = self.format(record)
            stream = self.stream
            stream.Write_shared((msg+self.terminator).encode(self.encoding))
            #self.flush()
        except Exception:
            self.handleError(record)

    def close(self):
        if self.stream:
            self.stream.Sync()
            self.stream.Close()
            self.stream = None

def initialize_logger(filename=None, file_level='info', stream=True, stream_level='info'):
    global file_handler
    global stream_handler
    addLoggingLevel('TRACE', TRACE)
    logger = logging.getLogger()
    logger.setLevel(min(get_log_level(file_level), get_log_level(stream_level))) # Lowest level that will be propagated to handlers
    logger.handlers = []
    if filename is not None:
        file_handler = MPIFileHandler(filename)
        file_handler.setLevel(get_log_level(file_level))
        logger.addHandler(file_handler)
    if stream:
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(get_log_level(stream_level))
        logger.addHandler(stream_handler)
    set_logging_prefix(MPI.COMM_WORLD.rank)
    if file_level == 'trace' or stream_level == 'trace':
        # Disable tracing of optimize for now
        add_trace_decorators(train=True, mpi=True, optimize=False)

def get_log_level(levelstr='info'):
    levelstr = levelstr.lower()
    return level_map[levelstr]

def set_logging_prefix(world_rank, parent_rank='-', process_rank='-', process_type='P'):
    global file_handler
    global stream_handler
    prefix = base_prefix.format(ptype=process_type, world=world_rank, parent=parent_rank, process=process_rank)
    formatter = ElapsedTimeFormatter('%(asctime)s ' + prefix + ' [%(levelname)s] %(message)s')
    if file_handler is not None:
        file_handler.setFormatter(formatter)
    if stream_handler is not None:
        stream_handler.setFormatter(formatter)

def add_trace_decorators(train=True, mpi=True, optimize=False):
    train_module_names = ['algo', 'data', 'model', 'optimizer']
    mpi_module_names = ['manager', 'process', 'single_process']
    optimize_module_names = ['coordinator', 'process_block']

    if train:
        for mod_name in train_module_names:
            mod = importlib.import_module('..train.' + mod_name, package=__name__)
            _decorate_module(mod)

    if mpi:
        for mod_name in mpi_module_names:
            mod = importlib.import_module('..mpi.' + mod_name, package=__name__)
            _decorate_module(mod)

    #if optimize:
    #    for mod_name in optimize_module_names:
    #        mod = importlib.import_module('..optimize.' + mod_name, package=__name__)
    #        _decorate_module(mod)

def _decorate_module(mod):
    # Module-level functions
    tl_func_list = inspect.getmembers(mod, inspect.isfunction)
    for name, func in tl_func_list:
        skip = False
        for regex in skip_trace:
            if re.search(regex, func.__qualname__) is not None:
                skip = True
                break
        if skip:
            logging.log(TRACE, "Skipping tracing of {}".format(func.__qualname__))
        else:
            setattr(mod, name, _trace(func))
    # Class functions
    cls_list = inspect.getmembers(mod, inspect.isclass)
    for _, clazz in cls_list:
        cls_func_list = inspect.getmembers(clazz, lambda x: inspect.isfunction(x) or inspect.ismethod(x))
        logging.log(TRACE, "Class {} function list {}".format(clazz.__name__, cls_func_list))
        for fname, func in cls_func_list:
            skip = False
            for regex in skip_trace:
                if re.search(regex, func.__qualname__) is not None:
                    skip = True
                    break
            if skip:
                logging.log(TRACE, "Skipping tracing of {}".format(func.__qualname__))
            else:
                setattr(clazz, fname, _trace(func))


skip_trace = [
    'Monitor.*',
    'Timeline.*',
    'Data.is_numpy_array',
    'MPIProcess.lookup_mpi_tag',
    'Thread.*'
]

def _trace(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        def str_type(obj):
            if type(obj) == list:
                return 'list[{}]'.format(len(obj))
            elif isinstance(obj, np.ndarray):
                return 'ndarray[{}]'.format(obj.shape)
            else:
                return str(obj)
        real_args = args[1:] if '.' in function.__qualname__ else args
        print_args = [str_type(arg) for arg in real_args]
        print_kwargs = ['{}={}'.format(k, str_type(v)) for k, v in kwargs.items()]
        params = ', '.join(print_args + print_kwargs)
        logging.log(TRACE, '>>> {}.{}({})'.format(function.__module__, function.__qualname__, params))

        ret_val = function(*args, **kwargs)

        logging.log(TRACE, '<<< {}.{}'.format(function.__module__, function.__qualname__))
        return ret_val

    return wrapper



### Utilities for mpi_learn module
import os
import sys
import numpy as np
import logging

class Error(Exception):
    pass

def weights_from_shapes(weights_shapes):
    """Returns a list of numpy arrays representing the NN architecture"""
    return [ np.zeros( shape, dtype=np.float32 ) for shape in weights_shapes ]

def shapes_from_weights(weights):
    """Returns a list of tuples indicating the array shape of each layer of the NN"""
    return [ w.shape for w in weights ]

def opt_tag_lookup(tag):
    """
    Gets the integer corresponding to the given tag string
    """
    tags = {
            'json':1,
            'result':2,
            'mbuilder':3,
            'params':4
            }
    return tags.get(tag, 0)

def import_keras(tries=10):
    """There is an issue when multiple processes import Keras simultaneously --
        the file .keras/keras.json is sometimes not read correctly.  
        as a workaround, just try several times to import keras."""
    for try_num in range(tries):
        try:
            stderr = sys.stderr
            sys.stderr = open(os.devnull, 'w')
            import tensorflow.keras as keras
            sys.stderr = stderr
            return
        except ValueError:
            logging.warning("Unable to import keras. Trying again: {0:d}".format(try_num))
            from time import sleep
            sleep(0.1)
    logging.error("Failed to import keras!")

def load_model(filename=None, model=None, weights_file=None, custom_objects={}):
    """Loads model architecture from JSON and instantiates the model.
        filename: path to JSON file specifying model architecture
        model:    (or) a Keras model to be cloned
        weights_file: path to HDF5 file containing model weights
	custom_objects: A Dictionary of custom classes used in the model keyed by name"""
    import_keras()
    from tensorflow.keras.models import model_from_json, clone_model
    if filename is not None:
        with open( filename ) as arch_f:
            json_str = arch_f.readline()
            new_model = model_from_json( json_str, custom_objects=custom_objects) 
    if model is not None:
        new_model = clone_model(model)
    if weights_file is not None:
        new_model.load_weights( weights_file )
    return new_model




### This script downloads the cifar10 dataset, unpacks it, splits it into four pieces, and saves 
# each piece in a separate h5 file.

from numpy import array_split
from keras.datasets import cifar10
from keras.utils import np_utils
from keras import backend as K
import h5py
import sys

(X_train, Y_train), (X_test, Y_test) = cifar10.load_data()

img_rows = 32
img_cols = 32
if K.image_dim_ordering() == 'th':
    X_train = X_train.reshape(X_train.shape[0], 3, img_rows, img_cols)
    X_test = X_test.reshape(X_test.shape[0], 3, img_rows, img_cols)
    input_shape = (3, img_rows, img_cols)
else:
    X_train = X_train.reshape(X_train.shape[0], img_rows, img_cols, 3)
    X_test = X_test.reshape(X_test.shape[0], img_rows, img_cols, 3)
    input_shape = (img_rows, img_cols, 3)

num_train_pieces = int(sys.argv[1]) if len(sys.argv)>1 else 24
num_test_pieces = int(sys.argv[2]) if len(sys.argv)>1 else 4
split_X_train = [ X.astype('float32') / 255 for X in array_split(X_train, num_train_pieces) ]
split_Y_train = [ np_utils.to_categorical(Y,10) for Y in array_split(Y_train, num_train_pieces) ]
split_X_test = [ X.astype('float32') / 255 for X in array_split(X_test, num_test_pieces) ]
split_Y_test = [ np_utils.to_categorical(Y,10) for Y in array_split(Y_test, num_test_pieces) ]

train_list = []
for i in range(num_train_pieces):
    train_name = "cifar10_train_%d.h5" % i
    train_list.append(train_name+"\n")
    train_outfile = h5py.File( train_name, 'w' )
    train_outfile.create_dataset( "features", data=split_X_train[i] )
    train_outfile.create_dataset( "labels", data=split_Y_train[i] )
    train_outfile.close()
with open('train_cifar10.list', 'w') as train_list_file:
    for f in train_list:
        train_list_file.write(f)

test_list = []
for i in range(num_test_pieces):
    test_name = "cifar10_test_%d.h5" % i
    test_list.append(test_name+"\n")
    test_outfile = h5py.File( test_name, 'w' )
    test_outfile.create_dataset( "features", data=split_X_test[i] )
    test_outfile.create_dataset( "labels", data=split_Y_test[i] )
    test_outfile.close()
with open('test_cifar10.list', 'w') as test_list_file:
    for f in test_list:
        test_list_file.write(f)



import numpy as np
from skopt.space import Real, Integer
from copy import deepcopy
import logging

class GA(object):
    """
    A genetic algorithm optimizer that gives the steps of evolutionary optimization.
    
    Parameters
    
    dimensions [list, shape=(n_dims,)]
        List of search space dimensions. Each search dimension is an instance of a
        'Dimension' object ('Real' or 'Integer')
    
    the argument num_iterations in ask specifies the number of generations
    """
    
    nonuniformityMutationConstant = 3
    blank = -1
    
    def __init__(self, dimensions, populationSize):
        self.paramRanges = dimensions
        self.numParams = len(self.paramRanges)
        self.populationSize = populationSize
        self.numElite = populationSize // 5
        self.fitness = np.array([self.blank]*populationSize)
        
        population = []
        for i in range(0, populationSize):
            chromosome = []
            for j in range(0, self.numParams):
                gene = float(self.paramRanges[j].rvs()[0])
                chromosome.append(gene)
            population.append(chromosome)
        self.population = population
    
    def setGenerations(self, maxGenerations):
        self.maxGenerations = maxGenerations
    
    def ask(self):
        rounded = deepcopy(self.population)
        for i in range(len(self.population)):
            for j in range(len(self.paramRanges)):
                if isinstance(self.paramRanges[j], Integer):
                    rounded[i][j] = int(rounded[i][j])
        logging.debug('Asked for: {}'.format(rounded))
        return rounded
    
    # needs params/results to have length populationSize
    def tell(self, params, results, generation):
        searchIn = self.ask()
        for i in range(len(searchIn)):
            for j in range(len(params)):
                if params[j] == searchIn[i] and self.fitness[i] == self.blank:
                    logging.debug('Found told params')
                    self.fitness[i] = 1/results[j]
        logging.debug('Fitness: {}'.format(self.fitness))
        if self.blank not in self.fitness:
            logging.debug('Population before stepping: {}'.format(self.population))
            self.step(generation)
            logging.debug('Population after stepping: {}'.format(self.population))
        
        loss = 1
        
        bestInd = self.fitness.argsort()[-1:][::-1][0]
        if self.fitness[bestInd] is not None:
                loss = 1 / self.fitness[bestInd]
        return self.population[bestInd], loss
        
    
    def step(self, generation):
        logging.debug('Stepping in generation {} with fitness: {}'.format(generation, self.fitness))
        children = []
        newFitness = [self.blank]*self.populationSize
        
        # elitism: copy fittest organisms
        eliteIndices = self.fitness.argsort()[-self.numElite:][::-1]
        for i in range(len(eliteIndices)):
            children.append(self.population[eliteIndices[i]])
        
        i = self.numElite
        while i < self.populationSize:
            p1 = self.tournamentSelect()
            p2 = self.tournamentSelect()
            children.append(self.mutate(self.crossover(p1, p2), generation))
            i += 1
        
        self.population = children
        self.fitness = np.array(newFitness)
    
    def tournamentSelect(self):
        arr = list(range(self.populationSize))
        ind1 = np.random.randint(self.populationSize)
        arr.remove(ind1)
        ind2 = np.random.choice(arr)
        
        if self.fitness[ind1] > self.fitness[ind2]:
            return ind1
        else:
            return ind2
    
    def crossover(self, p1, p2):
        moreFit = p1
        lessFit = p2
        if self.fitness[p2] > self.fitness[p1]:
            moreFit = p2
            lessFit = p1
        
        child = []
        for i in range(self.numParams):
            gene = np.random.random() * (self.population[moreFit][i] - self.population[lessFit][i]) + self.population[moreFit][i]
            gene = self.capGene(gene, i)
            child.append(gene)
        
        return child
    
    def capGene(self, gene, i):
        if gene < self.paramRanges[i].low:
            gene = float(self.paramRanges[i].low)
        elif gene > self.paramRanges[i].high:
            gene = float(self.paramRanges[i].high)
        return gene
    
    def mutate(self, chromosomeOrig, generation):
        chromosome = list(chromosomeOrig)
        for i in range(self.numParams):
            mutation = 1 - np.power(np.random.random(), np.power(1 - generation/self.maxGenerations, self.nonuniformityMutationConstant))
            
            if np.random.random() > 0.5:
                mutation *= self.paramRanges[i].high - chromosome[i]
            else:
                mutation *= -(chromosome[i] - self.paramRanges[i].low)
            chromosome[i] += mutation
            chromosome[i] = self.capGene(chromosome[i], i)
        return chromosome
    


import os
import psutil
import time
import json
from functools import wraps
from mpi4py import MPI

def timeline(original_function=None, category=None, **decorator_kwargs):
    """Decorates a function. Can be called with or without additional arguments. Name of event is the name of decorated function.
        Params:
          category: Optional category of event (useful to show/hide category of events in timeline viewer)
          decorator_kwargs: Additional arguments
    """
    def _decorate(function):
        @wraps(function)
        def wrapped_function(*args, **kwargs):
            Timeline.begin(function.__name__, category, **decorator_kwargs)
            if len(args) == 1 and not kwargs and callable(args[0]):
                ret_val = function()(args[0])
            else:
                ret_val = function(*args, **kwargs)
            Timeline.end(function.__name__, category, **decorator_kwargs)
            return ret_val

        return wrapped_function

    if original_function:
        return _decorate(original_function)

    return _decorate

class Timeline(object):
    """ Class that traces events to display on timeline """
    _enabled = False
    _events = []
    _process_name = os.getpid()
    _process_file = None
    _flush_every = 100

    @classmethod
    def _record(cls, event_name, type, category=None, **kwargs):
        if cls._enabled:
            ts = int(round(time.time() * 1000000))
            if category is None: category = "TRACE"
            tid = kwargs.get("tid", "Main")
            
            event = {"name": event_name, "cat": category, "ph": type, "pid": cls._process_name, "tid": tid, "ts": ts, "args": kwargs}
            
            cls._events.append(event)

            # Write events to process-specific file (in case collect() is never called)
            if cls._flush_every > 0 and len(cls._events) % cls._flush_every == 0:
                with open(cls._process_file, 'a+') as timeline_file:
                    timeline_file.write(",\n".join(map(json.dumps, cls._events[-cls._flush_every:])))

    @classmethod
    def begin(cls, name, category=None, **kwargs):
        """Marks the beginning of an event.
            Params:
              name: Name of event
              category: Optional category of event (useful to show/hide category of events in trace viewer)
              kwargs: Additional arguments
        """
        cls._record(name, "B", category, **kwargs)
    
    @classmethod
    def end(cls, name, category=None, **kwargs):
        """Marks the end of an event.
            Params:
              name: Name of event
              category: Optional category of event (useful to show/hide category of events in trace viewer)
              kwargs: Additional arguments
        """
        cls._record(name, "E", category, **kwargs)

    @classmethod
    def enable(cls, flush_file=None, flush_every=100):
        """Enables collection of timeline events.
            Params:
              flush_file: name of per-process timeline file to temporarily write events to, if None, 'str(os.getpid()) + _timeline.json' is used
              flush_every: Number of events to collect before flushing them to temporary file. 0 disables writing.
        """
        cls._enabled = True
        cls._process_file = flush_file or str(os.getpid()) + "_timeline.json"
        cls._flush_every = flush_every

    @classmethod
    def set_process_name(cls, process_name):
        """Sets a user-defined name for the process in timeline, as opposed to using just pid of process
            Params:
              process_name: Name to use
        """
        cls._process_name = process_name

    @classmethod
    def collect(cls, file_name=None, clean=False, comm=None):
        """Collects events from processes and writes them to a timeline file.
            Params:
              file_name: name of timeline file to write events to, if None, 'nnlo_timeline.json' is used
              clean: If True, removes process-specific temporary timeline files
              comm: MPI communicator to use, if None, MPI.COMM_WORLD is used 
        """
        if not cls._enabled:
            return

        comm = comm or MPI.COMM_WORLD
        all_events = comm.gather(cls._events, root=0)

        if (comm.Get_rank() == 0):
            with open(file_name or "nnlo_timeline.json", 'w+') as master_file:
                master_file.write("[\n")
                flat_events = []
                for events in all_events:
                    flat_events.append(",\n".join(map(json.dumps, events)))
                master_file.write(",\n".join(flat_events))
                master_file.write("\n]\n")

        if clean and os.path.isfile(cls._process_file):
            os.remove(cls._process_file)



### Optimizers used to update master process weights

import numpy as np
import copy
import pickle
import os
import re
import logging

class Optimizer(object):
    """Base class for optimization algorithms.
        Currently doesn't do anything."""

    def __init__(self):
        pass

    def reset(self):
        pass

    def apply_update(self, weights, gradient):
        raise NotImplementedError

    def save(self, fn = None):
        if fn is None:
            fn = 'master-opt-{}.algo'.format( os.getpid())
        d= open(fn,'wb')
        pickle.dump(self, d)
        d.close()
        logging.info("Saved state to %s", fn)

    def load(self, fn = 'algo_.pkl'):
        if not fn.endswith('.algo'):
            fn = fn + '.algo'
        try:
            d = open(fn, 'rb')
            new_self = pickle.load( d )
            d.close()
        except:
            new_self = None
        return new_self

class MultiOptimizer(Optimizer):
    def __init__(self, opt, s):
        self.opts = [copy.deepcopy(opt) for i in range(s)]

    def reset(self):
        for o in self.opts:
            o.reset()

    def apply_update(self, weights, gradient):
        r = []
        for o,w,g in zip(self.opts, weights, gradient):
            r.append( o.apply_update(w,g) )
        return r

class VanillaSGD(Optimizer):
    """Stochastic gradient descent with no extra frills.
          learning_rate: learning rate parameter for SGD"""

    def __init__(self, learning_rate=0.01):
        super(VanillaSGD, self).__init__()
        self.learning_rate = learning_rate

    def apply_update(self, weights, gradient):
        """Move weights in the direction of the gradient, by the amount of the
            learning rate."""
        new_weights = []
        for w, g in zip(weights, gradient):
            if type(w) == list:
                new_weights.append( [] )
                for ww, gg in zip(w,g):
                    new_weights[-1].append( np.subtract( ww, self.learning_rate*gg) )
            else:
                new_weights.append(np.subtract(w, self.learning_rate*g))
        return new_weights

class RunningAverageOptimizer(Optimizer):
    """Base class for AdaDelta, Adam, and RMSProp optimizers.
        rho (tunable parameter): decay constant used to compute running parameter averages
        epsilon (tunable parameter): small constant used to prevent division by zero
        running_g2: running average of the squared gradient, where squaring is done componentwise"""

    def __init__(self, rho=0.95, epsilon=1e-8):
        super(RunningAverageOptimizer, self).__init__()
        self.init_rho = rho
        self.init_epsilon = epsilon
        RunningAverageOptimizer.reset(self)

    def reset(self):
        super(RunningAverageOptimizer, self).reset()
        self.epsilon = self.init_epsilon
        self.rho = self.init_rho
        self.running_g2 = None

    def running_average_square_np(self, previous, update):
        """Computes and returns the running average of the square of a numpy array.
            previous (numpy array): value of the running average in the previous step
            update (numpy array): amount of the update"""
        try:
            new_contribution = (1-self.rho) * np.square(update)
            old_contribution = self.rho * previous
            return new_contribution + old_contribution
        except Exception as e:
            logging.error("FAILED TO COMPUTE THE RUNNING AVG SQUARE due to %s",str(e))
            logging.debug("rho %d",self.rho)
            logging.debug("min update %d",np.min(update))
            logging.debug("max update %d",np.max(update))
            logging.debug("min previous %d",np.min(previous))
            logging.debug("max previous %d",np.max(previous))
            return previous

    def running_average_square(self, previous, update):
        """Returns the running average of the square of a quantity.
            previous (list of numpy arrays): value of the running average in the previous step
            update (list of numpy arrays): amount of the update"""
        if previous == 0:
            previous = [ np.zeros_like(u) for u in update ]
        result = []
        for prev, up in zip(previous, update):
            result.append( self.running_average_square_np( prev, up ) )
        return result

    def sqrt_plus_epsilon(self, value):
        """Computes running RMS from the running average of squares.
            value: numpy array containing the running average of squares"""
        return np.sqrt( np.add(value, self.epsilon) )

class Adam(RunningAverageOptimizer):
    """Adam optimizer.
        Note that the beta_2 parameter is stored internally as 'rho'
        and "v" in the algorithm is called "running_g2"
        for consistency with the other running-average optimizers
        Attributes:
          learning_rate: base learning rate
          beta_1: decay rate for the first moment estimate
          m: running average of the first moment of the gradient
          t: time step
        """

    def __init__(self, learning_rate=0.001, beta_1=0.9, beta_2=0.999,
            epsilon=1e-8):
        super(Adam, self).__init__(rho=beta_2, epsilon=epsilon)
        self.init_learning_rate = learning_rate
        self.init_beta_1 = beta_1
        Adam.reset(self)

    def reset(self):
        super(Adam, self).reset()
        self.beta_1 = self.init_beta_1
        self.learning_rate = self.init_learning_rate
        self.t = 0
        self.m = None

    def running_average_np(self, previous, update):
        """Computes and returns the running average of a numpy array.
            Parameters are the same as those for running_average_square_np"""
        try:
            new_contribution = (1-self.beta_1) * update
            old_contribution = self.beta_1 * previous
            return new_contribution + old_contribution
        except Exception as e:
            logging.error("FAILED TO UPDATE THE RUNNING AVERAGE due to %s",str(e))
            logging.debug("beta_1 %d",self.beta_1)
            logging.debug("min update %d",np.min(update))
            logging.debug("max update %d",np.max(update))
            logging.debug("min previous %d",np.min(previous))
            logging.debug("max previous %d",np.max(previous))
            return previous


    def running_average(self, previous, update):
        """Returns the running average of the square of a quantity.
            Parameters are the same as those for running_average_square_np"""
        result = []
        for prev, up in zip(previous, update):
            result.append( self.running_average_np( prev, up ) )
        return result

    def apply_update(self, weights, gradient):
        """Update the running averages of the first and second moments
            of the gradient, and compute the update for this time step"""
        if self.running_g2 is None:
            self.running_g2 = [ np.zeros_like(g) for g in gradient ]
        if self.m is None:
            self.m = [ np.zeros_like(g) for g in gradient ]

        self.t += 1
        self.m = self.running_average( self.m, gradient )
        self.running_g2 = self.running_average_square( self.running_g2, gradient )
        alpha_t = self.learning_rate * (1 - self.rho**self.t)**(0.5) / (1 - self.beta_1**self.t)
        new_weights = []
        for w, g, g2 in zip(weights, self.m, self.running_g2):
            try:
                update = alpha_t * g / ( np.sqrt(g2) + self.epsilon )
            except Exception as e:
                logging.error("FAILED TO MAKE A WEIGHT UPDATE due to %s",str(e))
                logging.debug("alpha_t %d",alpha_t)
                logging.debug("beta_1 %d",self.beta_1)
                logging.debug("t %d",self.t)
                logging.debug("learning rate %d",self.learning_rate)
                logging.debug("rho %d",self.rho)
                logging.debug("epsilon %d",self.epsilon)
                logging.debug("min gradient %d",np.min( g ))
                logging.debug("max gradient %d",np.max( g ))
                logging.debug("min gradient 2 %d",np.min( g2 ))
                logging.debug("max gradient 2 %d",np.max( g2 ))
                try:
                    update = alpha_t * g / ( np.sqrt(g2) + self.epsilon )
                    try:
                        new_weights.append( w - update )
                    except:
                        logging.debug("no sub")
                except:
                    try:
                        update = g / ( np.sqrt(g2) + self.epsilon )
                        logging.debug("min b %d",np.min( update ))
                        logging.debug("max b %d",np.max( update ))
                        logging.debug("min |b| %d",np.min(np.fabs( update)))
                        #update *= alpha_t
                    except:
                        try:
                            update = 1./ ( np.sqrt(g2) + self.epsilon )
                        except:
                            try:
                                update = 1./ ( g2 + self.epsilon )
                            except:
                                pass
                update = 0
            new_weights.append( w - update )
        return new_weights

class AdaDelta(RunningAverageOptimizer):
    """ADADELTA adaptive learning rate method.
        running_dx2: running average of squared parameter updates
        """

    def __init__(self, rho=0.95, epsilon=1e-8):
        super(AdaDelta, self).__init__(rho, epsilon)
        AdaDelta.reset(self)

    def reset(self):
        super(AdaDelta, self).reset()
        self.running_dx2 = None

    def apply_update(self, weights, gradient):
        """Update the running averages of gradients and weight updates,
            and compute the Adadelta update for this step."""
        if self.running_g2 is None:
            self.running_g2 = [ np.zeros_like(g) for g in gradient ]
        if self.running_dx2 is None:
            self.running_dx2 = [ np.zeros_like(g) for g in gradient ]

        self.running_g2 = self.running_average_square( self.running_g2, gradient )
        new_weights = []
        updates = []
        for w, g, g2, dx2 in zip(weights, gradient, self.running_g2, self.running_dx2):
            update = np.multiply( np.divide( self.sqrt_plus_epsilon(dx2), self.sqrt_plus_epsilon(g2) ), g )
            new_weights.append( np.subtract( w, update ) )
            updates.append(update)
        self.running_dx2 = self.running_average_square( self.running_dx2, updates )
        return new_weights

class RMSProp(RunningAverageOptimizer):
    """RMSProp adaptive learning rate method.
        learning_rate: base learning rate, kept constant
        """

    def __init__(self, rho=0.9, epsilon=1e-8, learning_rate=0.001):
        super(RMSProp, self).__init__(rho, epsilon)
        self.init_learning_rate = learning_rate
        self.reset()

    def reset(self):
        super(RMSProp, self).reset()
        self.learning_rate = self.init_learning_rate

    def apply_update(self, weights, gradient):
        """Update the running averages of gradients,
            and compute the update for this step."""
        if self.running_g2 is None:
            self.running_g2 = [ np.zeros_like(g) for g in gradient ]

        self.running_g2 = self.running_average_square( self.running_g2, gradient )
        new_weights = []
        for w, g, g2 in zip(weights, gradient, self.running_g2):
            update = np.multiply( np.divide( self.learning_rate, self.sqrt_plus_epsilon(g2) ), g )
            new_weights.append( np.subtract( w, update ) )
        return new_weights

class TFOptimizer(Optimizer):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

        self.sess = None
        self.saver = None
        self.load_fn = None
        self.tf_optimizer = None

        self.reset()

    def reset(self):
        import tensorflow as tf

        if self.sess:
            self.sess.close() #free resources from previous execution

        self.sess = tf.Session()
        self.do_reset = True

    def save(self, fn='train_history'):
        fn = re.sub(r'\.algo$', '', fn)
        if not fn.startswith('./'):
            fn = './' + fn
        path = self.saver.save(self.sess, fn)
        logging.info("Saved state to %s", path)

    def setup_update(self, weights):
        import tensorflow as tf

        """Setup the tf computational graph. Should be run once for each model
            Receives the weights in order to know the shapes to use
        """
        self.gradient = [ tf.placeholder(dtype=tf.float32, shape=w.shape, name="gradient") for w in weights ]
        self.weights = [ tf.Variable(w, dtype=tf.float32, name="weights_{}".format(i)) for i, w in enumerate(weights) ]

        var_list = zip(self.gradient, self.weights)

        self.tf_time = tf.Variable(1, dtype=tf.float32, name="time")

        self.optimizer_op = self.tf_optimizer.apply_gradients(
            grads_and_vars=var_list,
            global_step=self.tf_time,
            name='optimizer_op' # We may need to create uniqie name
        )

        self.saver = tf.train.Saver(max_to_keep=None)

        if self.load_fn:
            self.saver.restore(self.sess, self.load_fn)
        else:
            self.sess.run(tf.global_variables_initializer())

    def load(self, fn='train_history'):
        load_fn = re.sub(r'\.algo$', '', fn)
        if os.path.isfile(load_fn + '.meta'):
            self.load_fn = load_fn
            self.do_reset = True
            return self
        else:
            return None

    def apply_update(self, weights, gradient):
        if self.do_reset:
            self.setup_update(weights)
            self.do_reset = False

        gradient_dict = {placeholder : value for placeholder, value in zip(self.gradient, gradient)}

        self.sess.run(self.optimizer_op, feed_dict=gradient_dict)
        res = self.sess.run(self.weights)

        return res

class GradientDescentTF(TFOptimizer):
    def __init__(self, learning_rate=0.01):
        super(GradientDescentTF, self).__init__(learning_rate=learning_rate)

    def setup_update(self, weights):
        import tensorflow as tf

        self.tf_optimizer = tf.train.GradientDescentOptimizer(
            learning_rate=self.learning_rate,
            use_locking=False,
            name='SGDMaster' # We may need to create uniqie name
        )

        super(GradientDescentTF, self).setup_update(weights)

class AdaDeltaTF(TFOptimizer):
    def __init__(self, learning_rate=0.001, rho=0.95, epsilon=1e-8):
        super(AdaDeltaTF, self).__init__(learning_rate=learning_rate, rho=rho, epsilon=epsilon)

    def setup_update(self, weights):
        import tensorflow as tf

        self.tf_optimizer = tf.train.AdadeltaOptimizer(
            learning_rate=self.learning_rate,
            rho=self.rho,
            epsilon=self.epsilon,
            use_locking=False,
            name='AdaDeltaMaster' # We may need to create uniqie name
        )

        super(AdaDeltaTF, self).setup_update(weights)

class RMSPropTF(TFOptimizer):
    def __init__(self, learning_rate=0.001, decay=0.9, momentum=0.0, epsilon=1e-10):
        super(RMSPropTF, self).__init__(learning_rate=learning_rate, decay=decay,
            momentum=momentum, epsilon=epsilon)

    def setup_update(self, weights):
        import tensorflow as tf

        self.tf_optimizer = tf.train.RMSPropOptimizer(
            learning_rate=self.learning_rate,
            decay=self.decay,
            momentum=self.momentum,
            epsilon=self.epsilon,
            use_locking=False,
            name='RMSPropMaster' # We may need to create uniqie name
        )

        super(RMSPropTF, self).setup_update(weights)

class AdamTF(TFOptimizer):
    def __init__(self, learning_rate=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-8):
        super(AdamTF, self).__init__(learning_rate=learning_rate,
            beta_1=beta_1, beta_2=beta_2, epsilon=epsilon)

    def setup_update(self, weights):
        import tensorflow as tf

        self.tf_optimizer = tf.train.AdamOptimizer(
            learning_rate=self.learning_rate,
            beta1=self.beta_1,
            beta2=self.beta_2,
            epsilon=self.epsilon,
            use_locking=False,
            name='AdamMaster' # We may need to create uniqie name
        )

        super(AdamTF, self).setup_update(weights)

class TorchOptimizer(Optimizer):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

        self.parameters = None
        self.torch_optimizer = None
        self.state = None

        self.reset()

    def reset(self):
        self.do_reset = True

    def apply_update(self, weights, gradient):
        import torch
        if self.do_reset:
            self.setup_update(weights)
            if self.state is not None:
                self.torch_optimizer.load_state_dict(self.state)
            self.do_reset = False

        for p, w, g in zip (self.parameters, weights, gradient):
            p.data.copy_(torch.from_numpy(w))
            p.grad.data.copy_(torch.from_numpy(g))

        self.torch_optimizer.step()
        return [i.data.cpu().numpy() for i in list(self.parameters)]

    def setup_update(self, weights):
        import torch
        if self.parameters is not None:
            for p in self.parameters:
                del p
        else:
            self.parameters = [None] * len(weights)
        for i, w in enumerate(weights):
            p = torch.from_numpy(w).cuda()
            g = torch.from_numpy(w).cuda()
            var = torch.autograd.Variable(p, requires_grad=True)
            var.grad = torch.autograd.Variable(g)
            self.parameters[i] = var

    def save(self, fn=None):
        if fn is None:
            fn = 'master-opt-{}.algo'.format( os.getpid())

        state = self.torch_optimizer.state_dict()
        with open(fn, 'wb') as out_file:
            pickle.dump(state, out_file)
        logging.info("Saved state to %s", fn)

    def load(self, fn = 'algo_.pkl'):
        if not fn.endswith('.algo'):
            fn = fn + '.algo'
        with open(fn, 'rb') as in_file:
            self.state = pickle.load(in_file)
        return self

class SGDTorch(TorchOptimizer):
    def __init__(self, lr=0.01, momentum=0, dampening=0, weight_decay=0, nesterov=False):
        super(SGDTorch, self).__init__(lr=lr, momentum=momentum,
            dampening=dampening, weight_decay=weight_decay, nesterov=nesterov)

    def setup_update(self, weights):
        super(SGDTorch, self).setup_update(weights)
        import torch.optim as topt
        self.torch_optimizer = topt.SGD(self.parameters, self.lr, self.momentum, self.dampening,
            self.weight_decay, self.nesterov)

class AdaDeltaTorch(TorchOptimizer):
    def __init__(self, lr=1.0, rho=0.9, eps=1e-06, weight_decay=0):
        super(AdaDeltaTorch, self).__init__(lr=lr, rho=rho, eps=eps, weight_decay=weight_decay)

    def setup_update(self, weights):
        super(AdaDeltaTorch, self).setup_update(weights)
        import torch.optim as topt
        self.torch_optimizer = topt.Adadelta(self.parameters, self.lr, self.rho, self.eps, self.weight_decay)

class RMSPropTorch(TorchOptimizer):
    def __init__(self, lr=0.01, alpha=0.99, eps=1e-08, weight_decay=0, momentum=0, centered=False):
        super(RMSPropTorch, self).__init__(lr=lr, alpha=alpha, eps=eps, weight_decay=weight_decay,
            momentum=momentum, centered=centered)

    def setup_update(self, weights):
        super(RMSPropTorch, self).setup_update(weights)
        import torch.optim as topt
        self.torch_optimizer = topt.RMSprop(self.parameters, self.lr, self.alpha, self.eps,
            self.weight_decay, self.momentum, self.centered)

class AdamTorch(TorchOptimizer):
    def __init__(self, lr=0.001, betas=(0.9, 0.999), eps=1e-08, weight_decay=0):
        super(AdamTorch, self).__init__(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)

    def setup_update(self, weights):
        super(AdamTorch, self).setup_update(weights)
        import torch.optim as topt
        self.torch_optimizer = topt.Adam(self.parameters, self.lr, self.betas, self.eps,
            self.weight_decay)

class GEM(Optimizer):
    """GEM optimizer
        learning_rate: base learning rate, kept constant
        momentum: momentum term, constant
        kappa: Proxy amplification. Experimental results show 2 is a good value.
        """

    def __init__(self, learning_rate=0.01, momentum=0.9, kappa=1.0):
        super(GEM, self).__init__()
        self.learning_rate = learning_rate
        self.momentum = momentum
        self.kappa = kappa
        self.epsilon = 10e-13

        self.central_variable_moment = None
        self.stale = None
        self.moment = None

        self.tensors_initialized = False

    def reset(self):
        self.tensors_initialized = False

    def init_tensors(self, weights):
        if not self.tensors_initialized:
            self.central_variable_moment = [ np.zeros_like(w) for w in weights ]
            self.stale = [ np.zeros_like(w) for w in weights ]
            self.moment = [ np.zeros_like(w) for w in weights ]

            self.tensors_initialized = True

    def begin_compute_update(self, cur_weights, new_weights):
        self.init_tensors(cur_weights)

        update = []

        for idx, (cur_w, new_w) in enumerate(zip(cur_weights, new_weights)):
            update.append( np.subtract( cur_w, new_w ))
            update[idx] *= -self.learning_rate
            # Update the states with the current gradient.
            self.moment[idx] *= self.momentum
            self.moment[idx] += update[idx]

        return update

    def gradient_energy_matching(self, gradient):
        Pi = [] # Pi tensors for all parameters.

        for idx, g in enumerate(gradient):
            proxy = self.kappa * np.abs(self.moment[idx])
            central_variable = np.abs(self.central_variable_moment[idx])
            update = np.abs(g + self.epsilon)
            pi = (proxy - central_variable) / update
            np.clip(pi, 0., 5., out=pi) # For numerical stability.
            Pi.append(pi)

        return Pi

    def compute_update(self, weights, gradient):
        for idx, w in enumerate(weights):
            self.central_variable_moment[idx] = (w - self.stale[idx])
            self.stale[idx] = np.copy(w)

        update = []

        pi = self.gradient_energy_matching(gradient)
        # Apply the scalars to the update.
        for idx, g in enumerate(gradient):
            update.append(np.multiply(g, pi[idx]))

        return update

    def apply_update(self, weights, update):
        """Add the update to the weights."""
        new_weights = []
        for w, u in zip(weights, update):
            new_weights.append(np.add(w, u))
        return new_weights

def get_optimizer(name):
    """Get optimizer class by string identifier"""
    lookup = {
            # Native optimizers
            'sgd':           VanillaSGD,
            'adadelta':      AdaDelta,
            'rmsprop':       RMSProp,
            'adam':          Adam,
            'gem':           GEM,
            # Wrappers around TF's optimizers
            'sgdtf':         GradientDescentTF,
            'adadeltatf':    AdaDeltaTF,
            'rmsproptf':     RMSPropTF,
            'adamtf':        AdamTF,
            # Wrappers arount Torch's optimizers
            'sgdtorch':      SGDTorch,
            'adadeltatorch': AdaDeltaTorch,
            'rmsproptorch':  RMSPropTorch,
            'adamtorch':     AdamTorch,
            }
    return lookup[name]

class OptimizerBuilder(object):
    """Builds a new Keras or Torch optimizer and optionally wraps it in horovod DistributedOptimizer."""

    def __init__(self, name, config=None, horovod_wrapper=False):
        self.name = name
        self.config = config
        if self.config is None:
            self.config = {}
        if self.name == 'sgd' and 'lr' not in self.config:
            logging.warning("Learning rate for SGD not set, using 1.0.")
            self.config['lr'] = 1.
        self.horovod_wrapper = horovod_wrapper

    def build(self):
        from tensorflow.keras.optimizers import deserialize
        opt_config = {'class_name': self.name, 'config': self.config}
        opt = deserialize(opt_config)
        if self.horovod_wrapper:
            import horovod.keras as hvd
            if hasattr(opt, 'lr'):
                opt.lr *= hvd.size()
            opt = hvd.DistributedOptimizer(opt)
        return opt

    def build_torch(self, model):
        import torch
        lookup = {
            'sgd':      torch.optim.SGD,
            'adadelta': torch.optim.Adadelta,
            'rmsprop':  torch.optim.RMSprop,
            'adam':     torch.optim.Adam
            }
        if self.name not in lookup:
            logging.warning("No optimizer '{}' found, using SGD instead".format(self.name))
            self.name = 'sgd'
        opt = lookup[self.name](model.parameters(), **self.config)
        if self.horovod_wrapper:
            import horovod.torch as hvd
            opt = hvd.DistributedOptimizer(opt, named_parameters=model.named_parameters())
        return opt



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections import defaultdict
try:
    import cPickle as pickle
except ImportError:
    import pickle
import tensorflow.keras as keras
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input
from tensorflow.keras import optimizers
from tensorflow.keras.optimizers import RMSprop,SGD
#from EcalEnergyGan import generator, discriminator
import numpy as np
import numpy.core.umath_tests as umath
import time
import socket
import os
import glob
import h5py
import logging

import keras.backend as K
from tensorflow.keras.models import Model, Sequential
from tensorflow.keras.layers import (Input, Dense, Reshape, Flatten, Lambda, merge,
                          Dropout, BatchNormalization, Activation, Embedding)
from tensorflow.keras.layers.advanced_activations import LeakyReLU
from tensorflow.keras.layers.convolutional import (UpSampling3D, Conv3D, ZeroPadding3D,
                                        AveragePooling3D)

from ..train.model import MPIModel, ModelBuilder
from .optimizer import OptimizerBuilder

kv2 = keras.__version__.startswith('2')

def hn():
    return socket.gethostname()


def weights(m):
    _weights_names = []
    for layer in m.layers:
        _weights_names += [ll.name for ll in layer.weights]
    _weights = m.get_weights()
    _disp = [(np.min(s),np.max(s),np.mean(s),np.std(s),s.shape,n) for s,n in zip(_weights,_weights_names)]

def weights_diff( m ,lap=True, init=False,label='', alert=None):#1000.):
    if (weights_diff.old_weights is None) or init:
        weights_diff.old_weights = m.get_weights()
        return
    _weights_names = []
    for layer in m.layers:
        _weights_names += [ll.name for ll in layer.weights]

    check_on_weight = m.get_weights()
    and_check_on_weight = weights_diff.old_weights
    ## make the diffs
    _diffs = [np.subtract(a,b) for (a,b) in zip(check_on_weight,and_check_on_weight)]
    _diffsN = [(np.min(s),np.max(s),np.mean(s),np.std(s),s.shape,n) for s,n in zip(_diffs,_weights_names)]
    for ii,dd in enumerate(_diffsN):
        if alert:
            if not any([abs(vv) > alert for vv in dd[:3]]):
                continue
    if lap:
        weights_diff.old_weights = m.get_weights()

weights_diff.old_weights = None

def _Conv3D(N,a,b,c,**args):
    if kv2:
        if 'border_mode'in args: args['padding'] = args.pop('border_mode')
        if 'init' in args: args['kernel_initializer'] = args.pop('init')
        if 'bias' in args: args['use_bias'] = args.pop('bias')
        return Conv3D(N,(a,b,c), **args)
    else:
        return Conv3D(N, a,b,c, **args)
def _BatchNormalization(**args):
    if kv2:
        m=0
        if 'mode' in args:
            m=args.pop('mode')
        if m==2:
            return StaticBatchNormalization(**args)
            #return BatchNormalization(**args)
        else:
            #args['scale'] = False
            #args['center'] = False
            return BatchNormalization(**args)
    else:
        return BatchNormalization(**args)

def _Dense(N,**args):
    if kv2:
        if 'init' in args: args['kernel_initializer'] = args.pop('init')
        return Dense(N,**args)
    else:
        return Dense(N,**args)

def _Model(**args):
    if kv2:
        args['outputs'] = args.pop('output')
        args['inputs'] = args.pop('input')
        return Model(**args)
    else:
        return Model(**args)
def discriminator(fixed_bn = False, discr_drop_out=0.2):

    image = Input(shape=( 25, 25, 25,1 ), name='image')

    bnm=2 if fixed_bn else 0
    f=(5,5,5)
    x = _Conv3D(32, 5, 5,5,border_mode='same',
               name='disc_c1')(image)
    x = LeakyReLU()(x)
    x = Dropout(discr_drop_out)(x)

    x = ZeroPadding3D((2, 2,2))(x)
    x = _Conv3D(8, 5, 5,5,border_mode='valid',
               name='disc_c2'
    )(x)
    x = LeakyReLU()(x)
    x = _BatchNormalization(name='disc_bn1',
                           mode=bnm,
    )(x)
    x = Dropout(discr_drop_out)(x)

    x = ZeroPadding3D((2, 2, 2))(x)
    x = _Conv3D(8, 5, 5,5,border_mode='valid',
               name='disc_c3'
)(x)
    x = LeakyReLU()(x)
    x = _BatchNormalization(name='disc_bn2',
                           #momentum = 0.00001
                           mode=bnm,
    )(x)
    x = Dropout(discr_drop_out)(x)

    x = ZeroPadding3D((1, 1, 1))(x)
    x = _Conv3D(8, 5, 5,5,border_mode='valid',
               name='disc_c4'
    )(x)
    x = LeakyReLU()(x)
    x = _BatchNormalization(name='disc_bn3',
                           mode=bnm,
    )(x)
    x = Dropout(discr_drop_out)(x)

    x = AveragePooling3D((2, 2, 2))(x)
    h = Flatten()(x)

    dnn = _Model(input=image, output=h, name='dnn')

    dnn_out = dnn(image)

    fake = _Dense(1, activation='sigmoid', name='classification')(dnn_out)
    aux = _Dense(1, activation='linear', name='energy')(dnn_out)
    ecal = Lambda(lambda x: K.sum(x, axis=(1, 2, 3)), name='sum_cell')(image)

    return _Model(output=[fake, aux, ecal], input=image, name='discriminator_model')

def generator(latent_size=200, return_intermediate=False, with_bn=True):

    latent = Input(shape=(latent_size, ))

    bnm=0
    x = _Dense(64 * 7* 7, init='glorot_normal',
              name='gen_dense1'
    )(latent)
    x = Reshape((7, 7,8, 8))(x)
    x = _Conv3D(64, 6, 6, 8, border_mode='same', init='he_uniform',
               name='gen_c1'
    )(x)
    x = LeakyReLU()(x)
    if with_bn:
        x = _BatchNormalization(name='gen_bn1',
                           mode=bnm
    )(x)
    x = UpSampling3D(size=(2, 2, 2))(x)

    x = ZeroPadding3D((2, 2, 0))(x)
    x = _Conv3D(6, 6, 5, 8, init='he_uniform',
               name='gen_c2'
    )(x)
    x = LeakyReLU()(x)
    if with_bn:
        x = _BatchNormalization(name='gen_bn2',
                           mode=bnm)(x)
    x = UpSampling3D(size=(2, 2, 3))(x)

    x = ZeroPadding3D((1,0,3))(x)
    x = _Conv3D(6, 3, 3, 8, init='he_uniform',
               name='gen_c3')(x)
    x = LeakyReLU()(x)

    x = _Conv3D(1, 2, 2, 2, bias=False, init='glorot_normal',
               name='gen_c4')(x)
    x = Activation('relu')(x)

    loc = _Model(input=latent, output=x)
    fake_image = loc(latent)
    _Model(input=[latent], output=fake_image)
    return _Model(input=[latent], output=fake_image, name='generator_model')

def get_sums(images):
    sumsx = np.squeeze(np.sum(images, axis=(2,3)))
    sumsy = np.squeeze(np.sum(images, axis=(1,3)))
    sumsz = np.squeeze(np.sum(images, axis=(1,2)))
    return sumsx, sumsy, sumsz

def get_moments(images, sumsx, sumsy, sumsz, totalE, m):
    ecal_size = 25
    totalE = np.squeeze(totalE)
    index = images.shape[0]
    momentX = np.zeros((index, m))
    momentY = np.zeros((index, m))
    momentZ = np.zeros((index, m))
    ECAL_midX = np.zeros(index)
    ECAL_midY = np.zeros(index)
    ECAL_midZ = np.zeros(index)
    if (totalE==0).any(): return momentX, momentY, momentZ
    for i in range(m):
      relativeIndices = np.tile(np.arange(ecal_size), (index,1))
      moments = np.power((relativeIndices.transpose()-ECAL_midX).transpose(), i+1)
      ECAL_momentX = umath.inner1d(sumsx, moments) /totalE
      if i==0: ECAL_midX = ECAL_momentX.transpose()
      momentX[:,i] = ECAL_momentX
    for i in range(m):
      relativeIndices = np.tile(np.arange(ecal_size), (index,1))
      moments = np.power((relativeIndices.transpose()-ECAL_midY).transpose(), i+1)
      ECAL_momentY = umath.inner1d(sumsy, moments) /totalE
      if i==0: ECAL_midY = ECAL_momentY.transpose()
      momentY[:,i]= ECAL_momentY
    for i in range(m):
      relativeIndices = np.tile(np.arange(ecal_size), (index,1))
      moments = np.power((relativeIndices.transpose()-ECAL_midZ).transpose(), i+1)
      ECAL_momentZ = umath.inner1d(sumsz, moments)/totalE
      if i==0: ECAL_midZ = ECAL_momentZ.transpose()
      momentZ[:,i]= ECAL_momentZ
    return momentX, momentY, momentZ

def load_sorted(sorted_path):
    sorted_files = sorted(glob.glob(sorted_path))

    energies = []
    srt = {}
    for f in sorted_files:
        file_name=f[f.find('sorted_'):-1]
        energy = int(''.join(list(filter(str.isdigit, file_name))[:-1]))*10
        energies.append(energy)
        srtfile = h5py.File(f,'r')
        srt["events" + str(energy)] = np.array(srtfile.get('ECAL'))
        srt["ep" + str(energy)] = np.array(srtfile.get('Target'))
    return energies, srt

def generate(g, index, latent, sampled_labels):
    noise = np.random.normal(0, 1, (index, latent))
    sampled_labels=np.expand_dims(sampled_labels, axis=1)
    gen_in = np.multiply(sampled_labels, noise)
    generated_images = g.predict(gen_in, verbose=False, batch_size=100)
    return generated_images



def metric(ganvar, g4var, energies, m):
   #calculate totale absolute errors on average moments+energies 
   ecal_size = 25
   metricp = 0
   metrice = 0
   for energy in energies:
     #Relative error on mean moment value for each moment and each axis
     if (g4var['moms_x'+str(energy)].all() > 0.0) :
        posx_error = (g4var['moms_x'+str(energy)] - ganvar['moms_x'+str(energy)])/g4var['moms_x'+str(energy)]
     else:
        posx_error = 1.
     if (g4var['moms_y'+str(energy)].all() > 0.0) :
        posy_error = (g4var['moms_y'+str(energy)] - ganvar['moms_y'+str(energy)])/g4var['moms_y'+str(energy)]
     else:
        posy_error = 1.
     if (g4var['moms_z'+str(energy)].all() > 0.0) :
        posz_error = (g4var['moms_z'+str(energy)] - ganvar['moms_z'+str(energy)])/g4var['moms_z'+str(energy)]
     else:
        posz_error = 1.
     #Taking absolute of errors and adding for each axis then scaling by 3
     pos_error = (np.absolute(posx_error) + np.absolute(posy_error) + np.absolute(posz_error))/3
     #Summing over moments and dividing for number of moments
     metricp += np.sum(pos_error)/m
     #Take profile along each axis and find mean along events
     eprofilex_error=0
     eprofiley_error=0
     eprofilez_error=0
     if (g4var['sumx'+str(energy)].all() > 0.0) :
        eprofilex_error = np.divide((g4var['sumx'+str(energy)] - ganvar['sumx'+str(energy)]), g4var['sumx'+str(energy)])
     else:
        eprofilex_error  =1.
        #raise ValueError('Image with NULL energy!')
     if (g4var['sumy'+str(energy)].all() > 0.0) :
        eprofiley_error = np.divide((g4var['sumy'+str(energy)] - ganvar['sumy'+str(energy)]), g4var['sumy'+str(energy)])
     else:
        eprofiley_error  =1.
        #raise ValueError('Image with NULL energy!')
     if (g4var['sumy'+str(energy)].all() > 0.0) :
        eprofilez_error = np.divide((g4var['sumz'+str(energy)] - ganvar['sumz'+str(energy)]), g4var['sumz'+str(energy)])
     else:
        eprofilez_error  =1.
        #raise ValueError('Image with NULL energy!')
     #Take absolute of error and mean for all events
     eprofilex_total = np.sum(np.absolute(eprofilex_error))/ecal_size
     eprofiley_total = np.sum(np.absolute(eprofiley_error))/ecal_size
     eprofilez_total = np.sum(np.absolute(eprofilez_error))/ecal_size

     metrice +=(eprofilex_total + eprofiley_total + eprofilez_total)/3
   tot = (metricp + metrice)/len(energies)
   return(tot)



def bit_flip(x, prob=0.05):
    """ flips a int array's values with some probability """
    x = np.array(x)
    selection = np.random.uniform(0, 1, x.shape) < prob
    x[selection] = 1 * np.logical_not(x[selection])
    return x


class StaticBatchNormalization(BatchNormalization):
    def call(self, inputs, training=None):
        return super(StaticBatchNormalization, self).call(inputs, training=False)

class GANModel(MPIModel):
    def __init__(self, **args):
        self.tell = args.get('tell',True)
        self.gen_bn = args.get('gen_bn',True)
        self._onepass = args.get('onepass',bool(int(os.environ.get('GANONEPASS',0))))
        self._reversedorder = args.get('reversedorder',bool(int(os.environ.get('GANREVERSED',0))))
        self._switchingloss = args.get('switchingloss',False)
        self._heavycheck = args.get('heavycheck',False)
        self._show_values = args.get('show_values',False)
        self._show_loss = args.get('show_loss', False)
        self._show_weights = False

        self.latent_size=args.get('latent_size',200)
        self.discr_drop_out=args.get('discr_drop_out',0.2)
        self.batch_size= None ## will be taken from the data that is passed on
        self.discr_loss_weights = [
            args.get('gen_weight',2),
            args.get('aux_weight',0.1),
            args.get('ecal_weight',0.1)
        ]
        
        self.with_fixed_disc = args.get('with_fixed_disc',True)
        self.assemble_models()
        self.recompiled = False
        self.checkpoint = args.get('checkpoint',int(os.environ.get('GANCHECKPOINT',0)))
        self.calculate_fom = args.get('calculate_fom',True)

        if self.tell:
            pass
        #if True:
        #    if self.with_fixed_disc: print ("the batch norm weights are fixed. heavey weight re-assigning")
        #    if self.checkpoint: print ("Checkpointing the model weigths after %d batch, based on the process id"%self.checkpoint)
        #    if self._onepass: print ("Training in one pass")
        #    if self._reversedorder: print ("will train generator first, then discriminator")
        #    if self._heavycheck: print("running heavy check on weight sanity")
        #    if self._show_values: print("showing the input values at each batch")
        #    if self._show_loss: print("showing the loss at each batch")
        #    if self._show_weights: print("showing weights statistics at each batch")

        MPIModel.__init__(self, models = [
            self.discriminator,
            #self.generator
            self.combined ## this is increasing a bit the amount of com by sending twice the discriminator
        ])

        
        ## counters
        self.g_cc = 0
        self.d_cc = 0
        self.p_cc = 0
        self.g_t = []
        self.d_t = []
        self.p_t = []


    def big_assemble_models(self):

        image = Input(shape=( 25, 25, 25,1 ), name='image')

        x = _Conv3D(32, 5, 5,5,border_mode='same')(image)
        x = LeakyReLU()(x)
        x = Dropout(discr_drop_out)(x)

        x = ZeroPadding3D((2, 2,2))(x)
        x = _Conv3D(8, 5, 5,5,border_mode='valid')(x)
        x = LeakyReLU()(x)
        x = BatchNormalization()(x)
        x = Dropout(discr_drop_out)(x)

        x = ZeroPadding3D((2, 2, 2))(x)
        x = Conv3D(8, 5, 5,5,border_mode='valid')(x)
        x = LeakyReLU()(x)
        x = BatchNormalization()(x)
        x = Dropout(discr_drop_out)(x)

        x = ZeroPadding3D((1, 1, 1))(x)
        x = Conv3D(8, 5, 5,5,border_mode='valid')(x)
        x = LeakyReLU()(x)
        x = BatchNormalization()(x)
        x = Dropout(discr_drop_out)(x)

        x = AveragePooling3D((2, 2, 2))(x)
        h = Flatten()(x)


        fake = Dense(1, activation='sigmoid', name='generation')(h)
        aux = Dense(1, activation='linear', name='auxiliary')(h)
        ecal = Lambda(lambda x: K.sum(x, axis=(1, 2, 3)), name='sum_cell')(image)

        self.discriminator = Model(output=[fake, aux, ecal], input=image, name='discriminator_model')

        latent = Input(shape=(self.latent_size, ))

        x = Dense(64 * 7* 7, init='he_uniform')(latent)
        x = Reshape((7, 7,8, 8))(x)
        x = Conv3D(64, 6, 6, 8, border_mode='same', init='he_uniform' )(x)
        x = LeakyReLU()(x)
        x = BatchNormalization()(x)
        x = UpSampling3D(size=(2, 2, 2))(x)

        x = ZeroPadding3D((2, 2, 0))(x)
        x = Conv3D(6, 6, 5, 8, init='he_uniform')(x)
        x = LeakyReLU()(x)
        x = BatchNormalization()(x)
        x = UpSampling3D(size=(2, 2, 3))(x)

        x = ZeroPadding3D((1,0,3))(x)
        x = Conv3D(6, 3, 3, 8, init='he_uniform')(x)
        x = LeakyReLU()(x)

        x = Conv3D(1, 2, 2, 2, bias=False, init='glorot_normal')(x)
        x = Activation('relu')(x)

        loc = Model(latent, x)
        #fake_image = loc(latent)
        self.generator = Model(input=latent, output=x, name='generator_model')

        c_fake, c_aux, c_ecal = self.discriminator(x)
        self.combined = Model(
            input = latent,
            output = [c_fake, c_aux, c_ecal],
            name='combined_model'
            )


    def ext_assemble_models(self):
        self.generator = generator(self.latent_size, with_bn = self.gen_bn)
        self.discriminator = discriminator(discr_drop_out = self.discr_drop_out)
        if self.with_fixed_disc:
            self.fixed_discriminator = discriminator(discr_drop_out = self.discr_drop_out, fixed_bn=True)
        latent = Input(shape=(self.latent_size, ), name='combined_z')
        fake_image = self.generator(latent)
        if self.with_fixed_disc:
            fake, aux, ecal = self.fixed_discriminator(fake_image)
        else:
            fake, aux, ecal = self.discriminator(fake_image)

        self.combined = Model(
            input=[latent],
            output=[fake, aux, ecal],
            name='combined_model'
        )

    def compile(self, **args):
        ## args are fully ignored here
        if 'optimizer' in args and isinstance(args['optimizer'], OptimizerBuilder):
            opt_builder = args['optimizer']
        else:
            opt_builder = None

        def make_opt(**args):
            if opt_builder:
                opt = opt_builder.build()
            else:
                ## there are things specified from outside mpi-learn
                lr = args.get('lr',0.0001)
                prop = args.get('prop',True) ## gets as the default
                if prop:
                    opt = RMSprop()    
                else:
                    opt = SGD(lr=lr)

            return opt

        self.generator.compile(
            optimizer=make_opt(**args),
            loss='binary_crossentropy') ## never  actually used for training

        self.discriminator.compile(
            optimizer=make_opt(**args),
            loss=['binary_crossentropy', 'mean_absolute_percentage_error', 'mean_absolute_percentage_error'],
            loss_weights=self.discr_loss_weights
        )

        if hasattr(self,'fixed_discriminator'):
            self.fixed_discriminator.trainable = False
        else:
            self.discriminator.trainable = False

        self.combined.compile(
            optimizer=make_opt(**args),
            loss=['binary_crossentropy', 'mean_absolute_percentage_error', 'mean_absolute_percentage_error'],
            loss_weights=self.discr_loss_weights
        )
        self.combined.metrics_names = self.discriminator.metrics_names

        
        if hasattr(self, 'calculate_fom'):
            self.energies, self.g4var = self.prepare_geant4_data()
         

    def assemble_models(self):
        self.ext_assemble_models()

    def batch_transform(self, x, y):
        root_fit = [0.0018, -0.023, 0.11, -0.28, 2.21]
        x_disc_real =x
        y_disc_real =y
        show_values = self._show_values
        def mm( label, t):
            pass

        if self.batch_size is None:
            ## fix me, maybe
            self.batch_size = x_disc_real.shape[0]


        noise = np.random.normal(0, 1, (self.batch_size, self.latent_size))
        sampled_energies = np.random.uniform(0.1, 5,(self.batch_size,1))
        generator_ip = np.multiply(sampled_energies, noise)
        if show_values: mm('energies',sampled_energies)
        ratio = np.polyval(root_fit, sampled_energies)
        if show_values: mm('ratios',ratio)
        ecal_ip = np.multiply(ratio, sampled_energies)
        if show_values: mm('estimated sum cells',ecal_ip)

        now = time.mktime(time.gmtime())
        generated_images = self.generator.predict(generator_ip)
        ecal_rip = np.squeeze(np.sum(generated_images, axis=(1, 2, 3)))
        if show_values: mm('generated sum cells',ecal_rip)

        norm_overflow = False
        apply_identify = False ## False was intended originally

        if norm_overflow and np.max( ecal_rip ) > 1000.:
            generated_images /= np.max( generated_images )
            ecal_rip = np.squeeze(np.sum(generated_images, axis=(1, 2, 3)))
            if show_values: mm('generated sum cells',ecal_rip)
        elif apply_identify:
            ecal_ip = ecal_rip

        done = time.mktime(time.gmtime())
        if self.p_cc:
            self.p_t.append( done - now )
        self.p_cc +=1

        ## need to bit flip the true labels too
        bf = bit_flip(y[0])
        y_disc_fake = [bit_flip(np.zeros(self.batch_size)), sampled_energies.reshape((-1,)), ecal_ip.reshape((-1,))]
        re_y = [bf, y_disc_real[1], y_disc_real[2]]

        if self._onepass:
            ## train the discriminator in one go
            bb_x  = np.concatenate( (x_disc_real, generated_images))
            bb_y = [np.concatenate((a,b)) for a,b in zip(re_y, y_disc_fake)]
            rng_state = np.random.get_state()
            np.random.shuffle( bb_x )
            np.random.set_state(rng_state)
            np.random.shuffle( bb_y[0] )
            np.random.set_state(rng_state)
            np.random.shuffle( bb_y[1] )
            np.random.set_state(rng_state)
            np.random.shuffle( bb_y[2] )

            X_for_disc = bb_x
            Y_for_disc = bb_y



        c_noise = np.random.normal(0, 1, (2*self.batch_size, self.latent_size))
        c_sampled_energies = np.random.uniform(0.1, 5, (2*self.batch_size,1 ))
        c_generator_ip = np.multiply(c_sampled_energies, c_noise)
        c_ratio = np.polyval(root_fit, c_sampled_energies)
        c_ecal_ip = np.multiply(c_ratio, c_sampled_energies)
        c_trick = np.ones(2*self.batch_size)

        X_for_combined = [c_generator_ip]
        Y_for_combined = [c_trick,c_sampled_energies.reshape((-1, 1)), c_ecal_ip]

        if self._onepass:
            return (X_for_disc,Y_for_disc,X_for_combined,Y_for_combined)
        else:
            def head(a):
                return [o[:self.batch_size] for o in a]
            def tail(a):
                return [o[self.batch_size:] for o in a]

            return ((x_disc_real,re_y),(generated_images,y_disc_fake), (head(X_for_combined),head(Y_for_combined)), (tail(X_for_combined),tail(Y_for_combined)))

    def test_on_batch(self,x, y, sample_weight=None):
        show_loss = self._show_loss
        if self._onepass:
            (X_for_disc,Y_for_disc,X_for_combined,Y_for_combined) = self.batch_transform(x,y)
            epoch_disc_loss = self.discriminator.test_on_batch(X_for_disc,Y_for_disc)
            epoch_gen_loss = self.combined.test_on_batch(X_for_combined,Y_for_combined)
        else:
            ((x_disc_real,re_y),(generated_images, y_disc_fake),(x_comb1,y_comb1),(x_comb2,y_comb2)) = self.batch_transform(x,y)
            real_disc_loss = self.discriminator.test_on_batch( x_disc_real,re_y )
            fake_disc_loss = self.discriminator.test_on_batch( generated_images, y_disc_fake)
            epoch_disc_loss = [(a + b) / 2 for a, b in zip(real_disc_loss, fake_disc_loss)]

            c_loss1= self.combined.test_on_batch( x_comb1,y_comb1 )
            c_loss2= self.combined.test_on_batch(x_comb2,y_comb2 )
            epoch_gen_loss = [(a + b) / 2 for a, b in zip(c_loss1,c_loss2)]




        return np.asarray([epoch_disc_loss, epoch_gen_loss])

    def train_on_batch(self, x, y,
                   sample_weight=None,
                   class_weight=None):

        with np.errstate( divide='raise', invalid='raise' , over='raise', under ='raise' ):
            if self._onepass:
                return self._onepass_train_on_batch(x,y,sample_weight,class_weight)
            else:
                return self._twopass_train_on_batch(x,y,sample_weight,class_weight)
    def _checkpoint(self):
        if self.checkpoint and (self.g_cc%self.checkpoint)==0:
            dest='%s/mpi_generator_%s_%s.h5'%(os.environ.get('GANCHECKPOINTLOC','.'),socket.gethostname(),os.getpid())
            logging.info("Saving generator to {} at {}".format(dest, self.g_cc))
            self.generator.save_weights(dest)        

    def _onepass_train_on_batch(self, x, y,
                   sample_weight=None,
                   class_weight=None):

        show_weights = self._show_weights
        show_loss = self._show_loss
        (X_for_disc,Y_for_disc,X_for_combined,Y_for_combined) = self.batch_transform(x,y)


        if self._heavycheck:
            on_weight = self.combined
            check_on_weight = on_weight.get_weights()
            if self._show_weights:
                weights( on_weight )
            weights_diff( on_weight , init=True)


        def _train_disc():
            self.discriminator.trainable = True
            now = time.mktime(time.gmtime())
            epoch_disc_loss = self.discriminator.train_on_batch(X_for_disc,Y_for_disc)
            done = time.mktime(time.gmtime())
            if self.d_cc:
                self.d_t.append( done - now )
            self.d_cc+=1
            if hasattr(self,'fixed_discriminator'):
                self.fixed_discriminator.set_weights( self.discriminator.get_weights())
            return epoch_disc_loss

        def _train_comb(noT=False):
            if hasattr(self,'fixed_discriminator'):
                self.fixed_discriminator.trainable = False
            else:
                self.discriminator.trainable = False
            now = time.mktime(time.gmtime())
            if noT:
                epoch_gen_loss = self.combined.test_on_batch(X_for_combined,Y_for_combined)
            else:
                epoch_gen_loss = self.combined.train_on_batch(X_for_combined,Y_for_combined)

            done = time.mktime(time.gmtime())
            if self.g_cc:
                self.g_t.append( done - now )
            self.g_cc+=1
            return epoch_gen_loss

        if self._reversedorder:
            epoch_gen_loss = _train_comb(noT=(self.g_cc==0))
            _pass = 'C-pass'
        else:
            epoch_disc_loss = _train_disc()
            _pass = 'D-pass'

        if self._heavycheck:
            if show_weights: weights( on_weight )
            weights_diff( on_weight , label=_pass)

        if self._reversedorder:
            epoch_disc_loss = _train_disc()
            _pass = 'D-pass'
        else:
            epoch_gen_loss = _train_comb()
            _pass = 'C-pass'

        if self._heavycheck:
            if show_weights: weights( on_weight )
            weights_diff( on_weight , label=_pass)

        if show_weights:
            weights( self.discriminator )
            weights( self.generator )
            weights( self.combined )


        self._checkpoint()

        return np.asarray([epoch_disc_loss, epoch_gen_loss])

    def _twopass_train_on_batch(self, x, y,
                   sample_weight=None,
                   class_weight=None):

        ((x_disc_real,re_y),(generated_images, y_disc_fake),(x_comb1,y_comb1),(x_comb2,y_comb2)) = self.batch_transform(x,y)

        show_loss = self._show_loss
        show_weights = self._show_weights
        self.discriminator.trainable = True

        if self._heavycheck:
            #on_weight = self.generator
            on_weight = self.combined
            check_on_weight = on_weight.get_weights()
            weights_names = []
            for l in on_weight.layers:
                weights_names += [ll.name for ll in l.weights]
            if self._show_weights:
                weights( on_weight )
            weights_diff( on_weight , init=True)

        now = time.mktime(time.gmtime())
        real_batch_loss = self.discriminator.train_on_batch(x_disc_real,re_y)

        if self._heavycheck:
            if hasattr(self,'fixed_discriminator'):
                self.fixed_discriminator.set_weights( self.discriminator.get_weights())
            if show_weights: weights( on_weight )
            weights_diff( on_weight , label='D-real')

        fake_batch_loss = self.discriminator.train_on_batch(generated_images, y_disc_fake)

        if hasattr(self,'fixed_discriminator'):
            ## pass things over
            self.fixed_discriminator.set_weights( self.discriminator.get_weights())
            self.fixed_discriminator.trainable = False
        else:
            self.discriminator.trainable = False

        if self._heavycheck:
            if show_weights: weights( on_weight )
            weights_diff( on_weight , label='D-fake')


        epoch_disc_loss = np.asarray([(a + b) / 2 for a, b in zip(real_batch_loss, fake_batch_loss)])
        done = time.mktime(time.gmtime())
        if self.d_cc:
            self.d_t.append( done - now )
        self.d_cc+=1

        if show_weights:
            weights( self.discriminator )
            weights( self.generator )
            weights( self.combined )

        if self.g_cc>1 and len(self.g_t)%100==0:
            now = time.mktime(time.gmtime())

        if self.g_cc:
            self.g_t.append( done - now )
        c_loss1= self.combined.train_on_batch( x_comb1,y_comb1)
        if self._heavycheck:
            if show_weights: weights( on_weight )
            weights_diff( on_weight , label ='C-1')
        c_loss2= self.combined.train_on_batch( x_comb2,y_comb2)
        #c_loss2= c_loss1
        if self._heavycheck:
            if show_weights: weights( on_weight )
            weights_diff( on_weight , label='C-2')

        epoch_gen_loss = np.asarray([(a + b) / 2 for a, b in zip(c_loss1,c_loss2)])
        done = time.mktime(time.gmtime())
        if self.g_cc:
            self.g_t.append( done - now )
        self.g_cc+=1

        if self._heavycheck:
            and_check_on_weight = on_weight.get_weights()

            ## this contains a boolean whether all values of the tensors are equal :
            # [--, False, --] ===> weights have changed for one layer
            # [--, True, --] ===> weights are the same one layer
            if show_weights:
                checks = [np.all(np.equal(a,b)) for (a,b) in zip(check_on_weight,and_check_on_weight)]
                weights_have_changed = not all(checks)
                weights_are_all_equal = all(checks)
                logging.info("Weights are the same? {}".format(str(checks)))
                if weights_have_changed:
                    for iw,b in enumerate(checks):
                        if not b:
                            logging.info("{} This {}".format(iw,str(check_on_weight[iw].shape)))
                            logging.info("{}".format(np.ravel(check_on_weight[iw])[:10]))
                            logging.info("{} And that {}".format(iw,and_check_on_weight[iw].shape))
                            logging.info("{}".format(np.ravel(and_check_on_weight[iw])[:10]))
                else:
                    logging.info("weights are all identical")
                    logging.info("".format(str(np.ravel(and_check_on_weight[1])[:10])))
                    logging.info("".format(str(np.ravel(check_on_weight[1])[:10])))

        self._checkpoint()

        ## modify learning rate
        if self._switchingloss:
            switching_loss = (1.,1.)
            if False and not self.recompiled and epoch_disc_loss[0]<switching_loss[0] and epoch_gen_loss[0]<switching_loss[1]:
                ## go on
                logging.info("going for full sgd")
                self.recompiled = True
                self.compile( prop=False, lr=1.0)
                #K.set_value( self.discriminator.optimizer.lr, 1.0)
                #K.set_value( self.generator.optimizer.lr, 1.0)

            lr = K.get_value(self.discriminator.optimizer.lr)
            nlr = lr
            ths = [
                #(5.5, 1.0),
                #(10., 0.1),
                (1000., 0.01),
                (10000., 0.001) ,
            ]
            for th in sorted(ths, key = lambda o : o[0]):
                if epoch_disc_loss[0] < th[0]:
                    nlr = th[1]
                    break
            if abs(nlr-lr)/lr > 0.0001:
                logging.info("{}".format("#"*30))
                logging.info("swithcing lr {} to {}".format(lr, nlr))
                K.set_value( self.discriminator.optimizer.lr, nlr)
                logging.info("{}".format(K.get_value( self.discriminator.optimizer.lr )))
                K.set_value( self.combined.optimizer.lr, nlr)
                logging.info("{}".format(K.get_value( self.combined.optimizer.lr )))

        return np.asarray([epoch_disc_loss, epoch_gen_loss])

    def prepare_geant4_data(self, **args):
        total = 0
        host = os.environ.get('HOST',os.environ.get('HOSTNAME',socket.gethostname()))
        if 'daint' in host:
            sortedpath = '/scratch/snx3000/vlimant/3DGAN/Sorted/sorted_*.hdf5'
        elif 'titan' in host:
            sortedpath = '//ccs/proj/csc291/DATA/3DGAN/sorted/sorted_*.hdf5'
        else:
            sortedpath = '/data/shared/3DGAN/sorted/sorted_*.hdf5'

        m = 2  #number of moments
        var = {}
        energies, srtev = load_sorted(sortedpath)
        for energy in energies:
            var["nevents" + str(energy)]= srtev["ep" + str(energy)].shape[0]
            var["ep"+str(energy)] = srtev["ep" + str(energy)]
            var["ecal_sum"+ str(energy)] = np.sum(srtev["events" + str(energy)], axis = (1, 2, 3))
            sumsx_act, sumsy_act, sumsz_act = get_sums(srtev["events" + str(energy)])
            momentX_act, momentY_act, momentZ_act = get_moments(srtev["events" + str(energy)], sumsx_act, sumsy_act, sumsz_act, var['ecal_sum'+str(energy)], m)
            var['moms_x'+str(energy)]=  np.mean(momentX_act, axis=0)
            var['moms_y'+str(energy)]=  np.mean(momentY_act, axis=0)
            var['moms_z'+str(energy)]=  np.mean(momentZ_act, axis=0)
            var['sumx'+str(energy)], var['sumy'+str(energy)], var['sumz'+str(energy)] = np.mean(sumsx_act, axis=0), np.mean(sumsy_act, axis=0), np.mean(sumsz_act, axis=0)

        return energies, var

    def figure_of_merit(self, **args):
        delta_loss = np.abs(self.histories['discriminator_model']['val_classification_loss'][-1] - self.histories['combined_model']['val_classification_loss'][-1])
        return delta_loss
        
        if (not self.calculate_fom) :
            raise ValueError('FOM not enabled: No Geant4 data calculated')
        total = 0
        m = 2  #number of moments
        latent= self.latent_size
        energies = self.energies
        g4var = self.g4var 
        ganvar={}
        if not energies:
           raise ValueError('No sorted file found')
        for energy in energies:
            events_gan  = generate(self.generator, g4var["nevents" + str(energy)], latent, g4var["ep" + str(energy)]/100.)
            events_gan[events_gan < 1e-4] = 0.
            ecal_gan  = np.sum(events_gan, axis = (1, 2, 3))
            sumsx_gan, sumsy_gan, sumsz_gan  = get_sums(events_gan)
            momentX_gan, momentY_gan, momentZ_gan = get_moments(events_gan, sumsx_gan, sumsy_gan, sumsz_gan, ecal_gan, m)

            ganvar["moms_x"+ str(energy)]= np.mean(momentX_gan, axis=0)
            ganvar["moms_y"+ str(energy)]= np.mean(momentY_gan, axis=0)
            ganvar["moms_z"+ str(energy)]= np.mean(momentZ_gan, axis=0)
            ganvar['sumx'+str(energy)], ganvar['sumy'+str(energy)], ganvar['sumz'+str(energy)] = np.mean(sumsx_gan, axis=0), np.mean(sumsy_gan, axis=0), np.mean(sumsz_gan, axis=0)

        return metric(ganvar, g4var, energies, m)



class GANModelBuilder(ModelBuilder):
    def __init__(self, c, tf=False, weights=None):
        ModelBuilder.__init__(self, c )
        self.tf = tf
        self.weights = weights# the prefix for the weights
        self.model_parameters={}
        
    def get_backend_name(self):
        return 'tensorflow'
    
    def set_params(self , **args):
        for k,v in args.items():
            self.model_parameters[k] = v

    def build_model(self, local_session=False):
        m = GANModel(**self.model_parameters)
        if self.weights:
            #for mm,w in zip(m.models, self.weights):
                #if w: mm.load_weights( w )
            for (im,mm) in enumerate(m.models):
                w_file = 'm%d_%s'%( im, self.weights )
                logging.info("Restoring weights from {}".format(w_file))
                mm.load_weights( w_file )
        return m

class GANBuilder(object):
    def __init__(self, parameters):
        self.parameters = parameters

    def builder(self,*params):
        args = dict(zip([p.name for p in self.parameters],params))
        gmb = GANModelBuilder(None) ## will be set later
        gmb.set_params(**args)
        return gmb




### MPIWorker and MPIMaster classes 

import os,sys,json
import numpy as np
import socket
import time
import threading, queue
import logging

from mpi4py import MPI

from ..util.monitor import Monitor
from ..util.timeline import Timeline, timeline
from ..util.utils import Error, weights_from_shapes, shapes_from_weights
from ..util.logger import set_logging_prefix

### Classes ###

class MPIProcess(object):
    """Base class for processes that communicate with one another via MPI.  

       Attributes:
           verbose: display verbose output
           parent_comm: MPI intracommunicator used to communicate with this process's parent
           parent_rank (integer): rank of this node's parent in parent_comm
           rank (integer): rank of this node in parent_comm
           model: Keras model to train
           model_builder: ModelBuilder object specifying model
           algo: Algo object defining how to optimize model weights
           weights_shapes: list of tuples indicating the shape of each layer of the model
           weights: list of numpy arrays storing the last weights received from the parent
           update: latest update obtained from training
           data: Data object used to generate training or validation data
           time_step: for keeping track of time
           num_epochs: integer giving the number of epochs to train for
           stop_training: becomes true when it is time to stop training
    """

    def __init__(self, parent_comm, process_comm, parent_rank=None, num_epochs=1, data=None, algo=None,
            model_builder=None, verbose=False, monitor=False, custom_objects={},
            checkpoint=None, checkpoint_interval=5):
        """If the rank of the parent is given, initialize this process and immediately start 
            training. If no parent is indicated, training should be launched with train().
            
            Parameters:
              parent_comm: MPI intracommunicator used to communicate with parent
              parent_rank (integer): rank of this node's parent in parent_comm
              num_epochs: number of training epochs
              data: Data object used to generate training or validation data
              algo: Algo object used to configure the training process
              model_builder: ModelBuilder object specifying model
              verbose: whether to print verbose output
              monitor: whether to monitor CPU/GPU usage
        """
        self.parent_comm = parent_comm
        self.process_comm = process_comm
        self.parent_rank = parent_rank
        self.num_epochs = num_epochs
        self.data = data
        self.algo = algo
        self.model_builder = model_builder
        self.verbose = verbose
        self.histories = {}
        self.custom_objects = custom_objects

        self.update = None
        self.stop_training = False
        self.time_step = 0
        self._short_batches = int(os.environ.get('NNLO_SHORT_BATCHES',0))
        self._is_shadow = (self.process_comm is not None and self.process_comm.Get_rank()!=0)

        self.monitor = Monitor() if monitor else None

        process_type = self.__class__.__name__.replace('MPI', '')[0]
        set_logging_prefix(
            MPI.COMM_WORLD.Get_rank(),
            self.parent_comm.Get_rank() if self.parent_comm is not None else '-',
            self.process_comm.Get_rank() if self.process_comm is not None else '-',
            process_type
        )

        if self.process_comm is not None and self.process_comm.Get_size() > 1:
            if self.model_builder.get_backend_name() == 'pytorch':
                import horovod.torch as hvd
            else:
                import horovod.keras as hvd
            logging.debug("initializing horovod")
            self.process_comm.Barrier()
            hvd.init(comm=self.process_comm)
            self.process_comm.Barrier()
            self.algo.worker_optimizer_builder.horovod_wrapper = True
        
        self.rank = parent_comm.Get_rank() if parent_comm else 0
        self.ranks = "{0}:{1}:{2}".format(
            MPI.COMM_WORLD.Get_rank(),
            self.parent_comm.Get_rank() if self.parent_comm is not None else '-',
            self.process_comm.Get_rank() if self.process_comm is not None else '-')

        Timeline.set_process_name(type(self).__name__ + " " + self.ranks)

        self.epoch = 0
        self.checkpoint = checkpoint
        self.checkpoint_interval = checkpoint_interval
        if self.algo.restore and self.checkpoint is not None:
            try:
                with open(self.checkpoint + '.latest', 'r') as latest:
                    latest_file = latest.read().splitlines()[-1]
                epoch = int(latest_file.split('-')[-1])
            except:
                epoch = 0
                logging.error("Failed to restore epoch from checkpoint {}".format(self.checkpoint))
            if epoch < num_epochs:
                self.epoch = epoch
                self.num_epochs = num_epochs - epoch
                logging.info("Continuing training from epoch {} for {} epochs".format(self.epoch, self.num_epochs))

        self.build_model()
        if (self.parent_rank is not None and self.parent_comm is not None):
            self.bcast_weights( self.parent_comm )
        if (self.parent_rank is not None and self.parent_comm is not None) or (self.process_comm):
            self.train()

    def record_details(self, json_name=None, meta=None):
        """to be implemented by the specific process
        """
        pass

    def save_checkpoint(self):
        if self.checkpoint is not None and self.epoch % self.checkpoint_interval == 0:
            file_name = '{}-{}'.format(self.checkpoint, self.epoch)
            logging.info("Checkpointing to {}".format( file_name ))
            if self.model:
                self.model.save(file_name + '.model')
                
            if self.algo:
                self.algo.save(file_name + '.algo')

            with open(self.checkpoint + '.latest', 'w') as latest:
                latest.write(file_name)

    def history_key(self):
        #return str(self.rank)
        return self.ranks

    def update_monitor(self, perf):
        r= self.history_key()
        self.histories.setdefault(r,{})['monitor'] = perf
        
    def update_history(self, items):
        r= self.history_key()
        self.model.update_history( items, self.histories.setdefault(r,{}))

    def is_shadow(self, sync = False):
        """signals that the process is a sub-process and should not act normally"""
        if self.process_comm and sync:
            import inspect
            logging.debug("syncing on the process communicator from",inspect.stack()[1][3])
            self.process_comm.Barrier()
        return self._is_shadow

    def build_model(self, local_session=True):
        """Builds the Keras model and updates model-related attributes"""
        self.model = self.model_builder.build_model(local_session=local_session)
        self.weights = self.model.get_weights()
        self.compile_model()
        self.weights = self.model.get_weights()
        self.update = self.model.format_update()

    def check_sanity(self):
        """Throws an exception if any model attribute has not been set yet."""
        for par in ['model',
                    #'weights_shapes',
                    'weights','update']:
            if not hasattr(self, par) or getattr(self, par) is None:
                raise Error("%s not found!  Process %s does not seem to be set up correctly." % (par,self.ranks))

    def train(self):
        """To be implemented in derived classes"""
        raise NotImplementedError

    def compile_model(self):
        """Compile the model. Note that the compilation settings
            are relevant only for Workers because the Master updates
            its weights using an mpi_learn optimizer."""
        logging.debug("Compiling model")
        self.algo.compile_model( self.model )
        logging.debug("Compiled")
        
    def print_metrics(self, metrics):
        """Display metrics computed during training or validation"""
        self.model.print_metrics( metrics )

    def get_logs(self, metrics, val=False):
        """Get dictionary of logs computed during training.
            If val is True, appends 'val' to the beginning of each metric name"""
        MPI.COMM_WORLD.Abort()
        if val:
            return { 'val_'+name:np.asscalar(metric) for name, metric in 
                    zip( self.model.metrics_names(), metrics ) }
        else:
            return { name:np.asscalar(metric) for name, metric in 
                    zip( self.model.metrics_names(), metrics ) }

    @timeline(category="MPI")
    def do_send_sequence(self):
        """Actions to take when sending an update to parent:
            -Send the update (if the parent accepts it)
            -Sync time and model weights with parent"""
        if self.is_shadow():
            return            
        self.send_update(check_permission=True)
        self.time_step = self.recv_time_step()
        self.recv_weights()
        self.algo.set_worker_model_weights( self.model, self.weights )

    def apply_update(self):
        """Updates weights according to update received from worker process"""
        with np.errstate( divide='raise',
                          invalid='raise',
                          over='raise',
                          #under ='raise'## might not be too bad
                      ):
            self.weights = self.algo.apply_update( self.weights, self.update )
        
    ### MPI-related functions below ###

    # This dict associates message strings with integers to be passed as MPI tags.
    tag_lookup = {
            'any':          MPI.ANY_TAG,
            'train':          0,
            'exit':           1,
            'begin_weights':  2,
            'begin_update'  : 3,
            'time':           4,
            'bool':           5,
            'history':        6,
            'weights':        12,
            'update':         12,
            'begin_gem':      13,
            'update_gem':     14,
            }
    # This dict is for reverse tag lookups.
    inv_tag_lookup = { value:key for key,value in tag_lookup.items() }

    def lookup_mpi_tag( self, name, inv=False ):
        """Searches for the indicated name in the tag lookup table and returns it if found.
            Params:
              name: item to be looked up
              inv: boolean that is True if an inverse lookup should be performed (int --> string)"""
        if inv:
            lookup = self.inv_tag_lookup
        else:
            lookup = self.tag_lookup
        try:
            return lookup[name]
        except KeyError:
            logging.error("Not found in tag dictionary: {0} -- returning None".format(name))
            return None

    def recv(self, obj=None, tag=MPI.ANY_TAG, source=None, buffer=False, status=None, comm=None):
        """Wrapper around MPI.recv/Recv. Returns the received object.
            Params:
              obj: variable into which the received object should be placed
              tag: string indicating which MPI tag should be received
              source: integer rank of the message source.  Defaults to self.parent_rank
              buffer: True if the received object should be sent as a single-segment buffer
                (e.g. for numpy arrays) using MPI.Recv rather than MPI.recv
              status: MPI status object that is filled with received status information
              comm: MPI communicator to use.  Defaults to self.parent_comm"""
        if comm is None:
            comm = self.parent_comm
        if source is None:
            if self.parent_rank is None:
                raise Error("Attempting to receive %s from parent, but parent rank is None" % tag)
            source = self.parent_rank 
        tag_num = self.lookup_mpi_tag(tag)
        if tag == 'history':
            obj = comm.recv( source=source, tag=tag_num, status=status )
            return obj
        #if tag in ['bool','time']:
        #    comm.Recv(obj, source=source, tag=tag_num, status=status )
        #    return obj
        if buffer:
            if type(obj) == list:
                for o in obj:
                    if type(o) == list:
                        for sub_o in o:
                            comm.Recv( sub_o, source=source, tag=tag_num, status=status )
                    else:
                        comm.Recv( o, source=source, tag=tag_num, status=status )
            else:
                comm.Recv( obj, source=source, tag=tag_num, status=status )
            return obj
        else:
            obj = comm.recv( source=source, tag=tag_num, status=status )
            return obj

    def send(self, obj, tag, dest=None, buffer=False, comm=None):
        """Wrapper around MPI.send/Send.  Params:
             obj: object to send
             tag: string indicating which MPI tag to send
             dest: integer rank of the message destination.  Defaults to self.parent_rank
             buffer: True if the object should be sent as a single-segment buffer
                (e.g. for numpy arrays) using MPI.Send rather than MPI.send
             comm: MPI communicator to use.  Defaults to self.parent_comm"""
        if comm is None:
            comm = self.parent_comm
        if dest is None:
            if self.parent_rank is None:
                raise Error("Attempting to send %s to parent, but parent rank is None" % tag)
            dest = self.parent_rank
        tag_num = self.lookup_mpi_tag(tag)
        if tag in ['history']:
            comm.send( obj, dest=dest, tag=tag_num )
            return
        #if tag in ['time']:
        #    comm.Send( obj, dest=dest, tag=tag_num )
        #    return
        if buffer:
            if type(obj) == list:
                for o in obj:
                    if type(o) == list:
                        for sub_o in o:
                            comm.Send( sub_o, dest=dest, tag=tag_num )
                    else:
                        comm.Send( o, dest=dest, tag=tag_num )
            else:
                comm.Send( obj, dest=dest, tag=tag_num )
        else:
            if type(obj) == list:
                for o in obj:
                    if type(o) == list:
                        for sub_o in o:
                            comm.Send( sub_o, dest=dest, tag=tag_num )
                    else:
                        comm.Send( o, dest=dest, tag=tag_num )
            else:
                comm.send( obj, dest=dest, tag=tag_num )

    def bcast(self, obj, root=0, buffer=False, comm=None):
        """Wrapper around MPI.bcast/Bcast.  Returns the broadcasted object.
            Params: 
              obj: object to broadcast
              root: rank of the node to broadcast from
              buffer: True if the object should be sent as a single-segment buffer
                (e.g. for numpy arrays) using MPI.Bcast rather than MPI.bcast
              comm: MPI communicator to use.  Defaults to self.parent_comm"""
        if comm is None:
            comm = self.parent_comm
        if buffer:
            if type(obj) == list:
                for o in obj:
                    if type(o) == list:
                        for sub_o in o:
                            comm.Bcast( sub_o, root=root )
                    else:
                        comm.Bcast(o, root=root )
            else:
                comm.Bcast( obj, root=root )
        else:
            obj = comm.bcast( obj, root=root )
            return obj

    def send_exit_to_parent(self):
        if self.is_shadow( sync = True): return
        """Send exit tag to parent process, if parent process exists"""
        if self.parent_rank is not None:
            self.send( None, 'exit' )

    def send_history_to_parent(self):
        if self.is_shadow():return        
        """Send keras history or dict of keras histories"""
        if self.parent_rank is not None:
            self.send( obj=self.histories, tag='history' )

    def send_arrays(self, obj, expect_tag, tag, comm=None, dest=None, check_permission=False):
        """Send a list of numpy arrays to the process specified by comm (MPI communicator) 
            and dest (rank).  We first send expect_tag to tell the dest process that we 
            are sending several buffer objects, then send the objects layer by layer.
            Optionally check first to see if the update will be accepted by the master"""
        self.send( None, expect_tag, comm=comm, dest=dest )
        if check_permission:
            # To check permission we send the update's time stamp to the master.
            # Then we wait to receive the decision yes/no.
            self.send_time_step( comm=comm, dest=dest )
            decision = self.recv_bool( comm=comm, source=dest )
            if not decision: return
        for o in obj:
            if type(o) == list:
                for w in o:
                    self.send( w, tag, comm=comm, dest=dest, buffer=True )
            else:
                self.send( o, tag, comm=comm, dest=dest, buffer=True )

    def send_weights(self, comm=None, dest=None, check_permission=False):
        if self.is_shadow():return        
        """Send NN weights to the process specified by comm (MPI communicator) and dest (rank).
            Before sending the weights we first send the tag 'begin_weights'."""
        self.send_arrays( self.weights, expect_tag='begin_weights', tag='weights', 
                comm=comm, dest=dest, check_permission=check_permission )

    def send_update(self, comm=None, dest=None, check_permission=False):
        if self.is_shadow():return        
        """Send update to the process specified by comm (MPI communicator) and dest (rank).
            Before sending the update we first send the tag 'begin_update'"""
        self.send_arrays( self.update, expect_tag='begin_update', tag='update', 
                comm=comm, dest=dest, check_permission=check_permission )

    def send_time_step(self, comm=None, dest=None):
        if self.is_shadow():return        
        """Send the current time step"""
        self.send( obj=self.time_step, tag='time', dest=dest, comm=comm )

    def send_bool(self, obj, comm=None, dest=None):
        if self.is_shadow():return        
        self.send( obj=obj, tag='bool', dest=dest, comm=comm )

    def recv_arrays(self, obj, tag, comm=None, source=None, add_to_existing=False):
        """Receive a list of numpy arrays from the process specified by comm (MPI communicator) 
            and dest (rank).
              obj: list of destination arrays 
              tag: MPI tag accompanying the message
              add_to_existing: if true, add to existing object instead of replacing"""
        if add_to_existing:
            #TODO this needs some work
            tmp = weights_from_shapes( [ w.shape for w in obj ] )
            self.recv_arrays( tmp, tag=tag, comm=comm, source=source )
            for i in range(len(obj)):
                obj[i] += tmp[i]
            return
        for o in obj:
            if type(o) == list:
                for w in o:
                    self.recv( w, tag, comm=comm, source=source, buffer=True )
            else:
                self.recv( o, tag, comm=comm, source=source, buffer=True )

    def recv_weights(self, comm=None, source=None, add_to_existing=False):
        """Receive NN weights layer by layer from the process specified by comm and source"""
        if self.is_shadow():return        
        self.recv_arrays( self.weights, tag='weights', comm=comm, source=source,
                add_to_existing=add_to_existing )

    def recv_update(self, comm=None, source=None, add_to_existing=False):
        """Receive an update layer by layer from the process specified by comm and source.
            Add it to the current update if add_to_existing is True, 
            otherwise overwrite the current update"""
        if self.is_shadow():return        
        self.recv_arrays( self.update, tag='update', comm=comm, source=source,
                add_to_existing=add_to_existing )

    def recv_time_step(self, comm=None, source=None):
        """Receive the current time step"""
        if self.is_shadow():return
        return self.recv( tag='time', comm=comm, source=source )

    def recv_bool(self, comm=None, source=None):
        if self.is_shadow():return        
        return self.recv( tag='bool', comm=comm, source=source )

    def recv_exit_from_parent(self):
        ir = None
        if not self.is_shadow():
            ir = self.parent_comm.irecv( source=0, tag=self.lookup_mpi_tag('exit') )
        elif self.process_comm:
            ir = self.process_comm.irecv( source=0, tag=self.lookup_mpi_tag('exit') )
        return ir


    def bcast_weights(self, comm, root=0):
        """Broadcast weights shape and weights (layer by layer) 
            on communicator comm from the indicated root rank"""
        self.bcast( self.weights, comm=comm, root=root, buffer=True )


class MPIWorker(MPIProcess):
    """This class trains its NN model and exchanges weight updates with its parent."""

    def __init__(self, data, algo, model_builder, process_comm, parent_comm, parent_rank=None, 
            num_epochs=1, verbose=False, monitor=False, custom_objects={},
            checkpoint=None, checkpoint_interval=5):
        """Raises an exception if no parent rank is provided. Sets the number of epochs 
            using the argument provided, then calls the parent constructor"""
        info = "Creating MPIWorker with rank {0} and parent rank {1} on a communicator of size {2}"
        tell_comm = parent_comm if parent_comm is not None else process_comm
        if tell_comm: logging.info(info.format(tell_comm.Get_rank(),
                           parent_rank,
                           tell_comm.Get_size()))

        super(MPIWorker, self).__init__( parent_comm, process_comm, parent_rank,
                num_epochs=num_epochs, data=data, algo=algo, model_builder=model_builder,
                verbose=verbose, monitor=monitor, custom_objects=custom_objects,
                checkpoint=checkpoint, checkpoint_interval=checkpoint_interval )

    def build_model(self):
        """Builds the Keras model and updates model-related attributes"""

        self.validation_model = self.model_builder.build_model()
        self.algo.compile_model( self.validation_model )
        super(MPIWorker, self).build_model(local_session=False)

    def sync_with_parent(self):
        self.compute_update()
        if self.algo.mode == 'gem':
            self.do_gem_sequence()
        else:
            self.do_send_sequence()

    def notify_parent(self):
        """Notify parent that worker is ready to begin GEM sequence."""
        self.send( None, 'begin_gem' )

    def do_gem_sequence(self):
        """GEM sequence on worker:
         -Notify parent we are ready
         -Receive central variable (weights) from master
         -Compute the update with GEM
         -Send the update to master and apply it to own weights
        """
        if self.is_shadow(): return
        self.notify_parent()
        self.recv_weights()
        self.update = self.algo.compute_update_worker(self.weights, self.update)
        self.send_update()
        self.apply_update()
        self.algo.set_worker_model_weights(self.model, self.weights)

    def train(self):
        """Wait for the signal to train. Then train for num_epochs epochs.
            In each step, train on one batch of input data, then send the update to the master
            and wait to receive a new set of weights.  When done, send 'exit' signal to parent.
        """
        Timeline.begin("train")
        self.check_sanity()

        self.await_signal_from_parent()

        # periodically check this request to see if the parent has told us to stop training
        exit_request = self.recv_exit_from_parent()
        for epoch in range(1, self.num_epochs + 1):
            logging.info("Beginning epoch {:d}".format(self.epoch + epoch))
            Timeline.begin("epoch")
            if self.monitor:
                self.monitor.start_monitor()
            epoch_metrics = np.zeros((1,))
            i_batch = 0

            for i_batch, batch in enumerate(self.data.generate_data()):
                if self.process_comm:
                    ## broadcast the weights to all processes
                    ## alternative is to load the broadcast callback from horovod and call it here
                    self.bcast_weights( comm=self.process_comm )
                    if self.process_comm.Get_rank()!=0:
                        self.model.set_weights(self.weights)

                Timeline.begin("train_on_batch")
                train_metrics = self.model.train_on_batch( x=batch[0], y=batch[1] )
                Timeline.end("train_on_batch")
                if epoch_metrics.shape != train_metrics.shape:
                    epoch_metrics = np.zeros( train_metrics.shape)
                epoch_metrics += train_metrics
                if self.algo.should_sync():
                    self.sync_with_parent()
                if exit_request and exit_request.Test():
                    self.stop_training = True
                    if self.process_comm:
                        for r in range(1, self.process_comm.Get_size()):
                            ## propagate the exit signal to processes of this worker
                            self.send_exit_to_child(r, comm=self.process_comm)
                    logging.info("Received exit request from master")
                    break
                if self._short_batches and i_batch>self._short_batches: break

            if self.monitor:
                self.monitor.stop_monitor()
            epoch_metrics = epoch_metrics / float(i_batch+1)
            l = self.model.get_logs( epoch_metrics )
            self.update_history( l )
            Timeline.end("epoch")

            if self.stop_training:
                break

        logging.debug("Signing off")
        stats = None
        if self.monitor:
            stats = self.monitor.get_stats()
            self.update_monitor( stats )
        Timeline.end("train", monitoring_stats=stats)
        self.send_exit_to_parent()
        self.send_history_to_parent()
        self.data.finalize()

    @timeline
    def compute_update(self):
        """Compute the update from the new and old sets of model weights"""
        self.update = self.algo.compute_update( self.weights, self.model.get_weights() )

    def await_signal_from_parent(self):
        """Wait for 'train' signal from parent process"""
        if not self.is_shadow():
            tag = self.recv( tag='train' )
        if self.process_comm:
            if self.process_comm.Get_rank() == 0:
                for r in range(1,self.process_comm.Get_size()):
                    self.send( None, tag='train', comm=self.process_comm, dest=r)
            else:
                self.recv( tag='train', comm=self.process_comm )



class MPIMaster(MPIProcess):
    """This class sends model information to its worker processes and updates its model weights
        according to updates or weights received from the workers.
        
        Attributes:
          child_comm: MPI intracommunicator used to communicate with child processes
          has_parent: boolean indicating if this process has a parent process
          num_workers: integer giving the number of workers that work for this master
          best_val_loss: best validation loss computed so far during training
          running_workers: list of workers not yet done training
          waiting_workers_list: list of workers that sent updates and are now waiting
          num_sync_workers: number of worker updates to receive before performing an update
          update_tag: MPI tag to expect when workers send updates
          histories: model histories collected from workers
          epoch: current epoch number
    """

    def __init__(self, parent_comm, parent_rank=None, child_comm=None, 
            num_epochs=1, data=None, algo=None, model_builder=None, 
                 num_sync_workers=1, verbose=False, monitor=False, custom_objects={}, early_stopping=None, target_metric=None,
                 threaded_validation=False, checkpoint=None, checkpoint_interval=5):
        """Parameters:
              child_comm: MPI communicator used to contact children"""
        if child_comm is None:
            raise Error("MPIMaster initialized without child communicator")
        self.child_comm = child_comm
        self.has_parent = False
        if parent_rank is not None:
            self.has_parent = True
        self.best_val_loss = None
        self.target_metric = (target_metric if type(target_metric)==tuple else tuple(map(lambda s : float(s) if s.replace('.','').isdigit() else s, target_metric.split(',')))) if target_metric else None
        self.patience = (early_stopping if type(early_stopping)==tuple else tuple(map(lambda s : float(s) if s.replace('.','').isdigit() else s, early_stopping.split(',')))) if early_stopping else None
        ## a couple of example
        ##self.target_metric = ('val_acc','>', 0.97)
        #self.patience = ('val_loss','~<',4)
        #self.target_metric = ('combined_model:val_loss','<',1)#('val_acc','>', 0.97)
        #self.patience = ('combined_model:val_loss','~<',2)        
        self.num_workers = child_comm.Get_size() - 1 #all processes but one are workers
        self.num_sync_workers = num_sync_workers
        info = ("Creating MPIMaster with rank {0} and parent rank {1}. "
                "(Communicator size {2}, Child communicator size {3})")
        logging.info(info.format(parent_comm.Get_rank(),parent_rank,parent_comm.Get_size(), 
                child_comm.Get_size()))
        if self.num_sync_workers > 1:
            logging.info("Will wait for updates from {0:d} workers before synchronizing".format(self.num_sync_workers))
        self.threaded_validation = threaded_validation

        super(MPIMaster, self).__init__( parent_comm, process_comm=None,parent_rank=parent_rank, data=data, 
                algo=algo, model_builder=model_builder, num_epochs=num_epochs, 
                verbose=verbose, monitor=monitor, custom_objects=custom_objects,
                checkpoint=checkpoint, checkpoint_interval=checkpoint_interval )

    def decide_whether_to_sync(self):
        """Check whether enough workers have sent updates"""
        return ( len(self.waiting_workers_list) >= self.num_sync_workers )

    def is_synchronous(self):
        return self.num_sync_workers > 1

    def accept_update(self):
        """Returns true if the master should accept the latest worker's update, false otherwise"""
        return (not self.is_synchronous()) or self.algo.staleness == 0
        
    def sync_children(self):
        """Update model weights and signal all waiting workers to work again.
            Send our update to our parent, if we have one"""
        while self.waiting_workers_list:
            child = self.waiting_workers_list.pop()
            self.sync_child(child)

    def sync_child(self, child):
        self.send_time_step( dest=child, comm=self.child_comm )
        self.send_weights( dest=child, comm=self.child_comm )

    def sync_parent(self):
        if self.has_parent:
            self.do_send_sequence()
        else:
            self.time_step += 1 

    def do_update_sequence(self, source):
        """Update procedure:
         -Compute the staleness of the update and decide whether to accept it.
         -If we accept, we signal the worker and wait to receive the update.
         -After receiving the update, we determine whether to sync with the workers.
         -Finally we run validation if we have completed one epoch's worth of updates."""
        child_time = self.recv_time_step( source=source, comm=self.child_comm )
        self.algo.staleness = self.time_step - child_time
        accepted = self.accept_update()
        self.send_bool( accepted, dest=source, comm=self.child_comm )
        if accepted:
            self.recv_update( source=source, comm=self.child_comm, 
                    add_to_existing=self.is_synchronous() )
            self.waiting_workers_list.append(source)
            if self.decide_whether_to_sync():
                if self.algo.send_before_apply:
                    self.sync_parent()
                    self.sync_children()
                    self.apply_update()
                else:
                    self.apply_update()
                    self.sync_parent()
                    self.sync_children()
                self.update = self.model.format_update()
            if (self.algo.validate_every > 0 and self.time_step > 0):
                if (self.time_step % self.algo.validate_every == 0) or (self._short_batches and self.time_step%self._short_batches == 0):
                    self.validate(self.weights)
                    self.epoch += 1
                    self.save_checkpoint()
        else:
            self.sync_child(source)

    def do_gem_sequence(self, source):
        """Gradient enery matching procedure:
         -Send the current central variable (weights) to worker.
         -Wait to receive the update and apply it.
         -Finally we run validation if we have completed one epoch's worth of updates."""
        self.send_weights( dest=source, comm=self.child_comm )

    def do_gem_update_sequence(self, source):
        self.recv_update( source=source, comm=self.child_comm )
        self.apply_update()
        #Update model weights and signal all waiting workers to work again.
        self.time_step += 1
        if (self.algo.validate_every > 0 and self.time_step > 0):
            if (self.time_step % self.algo.validate_every == 0) or (self._short_batches and self.time_step%self._short_batches == 0):
                self.validate(self.weights)
                self.epoch += 1
                self.save_checkpoint()

    def do_worker_finish_sequence(self, worker_id):
        """Actions to take when a worker finishes training and returns its history"""
        self.histories.update( self.recv_history_from_child(worker_id) )
        #key = "%d_%d" % (self.rank, worker_id)
        #self.histories[key] = self.recv_history_from_child(worker_id)
        self.running_workers.remove(worker_id)
        self.num_sync_workers -= 1

    def process_message(self, status):
        """Extracts message source and tag from the MPI status object and processes the message. 
            Returns the tag of the message received.
            Possible messages are:
            -begin_update: worker is ready to send a new update
            -begin_gem: Worker needs central variable to start GEM
            -exit: worker is done training and will shut down
        """
        source = status.Get_source()
        tag = self.lookup_mpi_tag( status.Get_tag(), inv=True )
        if tag == 'begin_update':
            if self.algo.mode == 'gem':
                self.do_gem_update_sequence(source)
            else:
                self.do_update_sequence(source)
        elif tag == 'begin_gem':
            self.do_gem_sequence(source)
        elif tag == 'exit':
            self.do_worker_finish_sequence(source)
        else:
            raise ValueError("Tag %s not recognized" % tag)
        return tag

    def shut_down_workers(self):
        """Signal all running workers to shut down"""
        for worker_id in self.running_workers:
            logging.info("Signaling worker {0:d} to shut down".format(worker_id))
            self.send_exit_to_child( worker_id )

    def train(self):
        """Broadcasts model information to children and signals them to start training.
            Receive messages from workers and processes each message until training is done.
            When finished, signal the parent process that training is complete.
        """
        Timeline.begin("train")
        self.start_time = time.time()
        self.check_sanity()
        self.bcast_weights( comm=self.child_comm )
        self.signal_children()

        if self.threaded_validation:
            self.validation_queue = queue.Queue()
            self.validation_thread = threading.Thread(target=MPIMaster.validation_worker, args = (self, ))
            self.validation_thread.start()

        status = MPI.Status()
        self.running_workers = list(range(1, self.num_workers+1))
        self.waiting_workers_list = []
        
        while self.running_workers:
            self.recv_any_from_child(status)
            self.process_message( status )
            if (self.stop_training):
                self.shut_down_workers()
        logging.info("Done training")
        # If we did not finish the last epoch, validate one more time.
        # (this happens if the batch size does not divide the dataset size)
        if self.epoch < self.num_epochs or not self.histories.get(self.history_key(),None):
            epoch_logs = self.validate(self.weights)
        if self.threaded_validation:
            self.validation_queue.put(None)
            self.validation_queue.join()
            self.validation_thread.join()
        self.save_checkpoint()
        self.send_exit_to_parent()
        self.send_history_to_parent()
        self.data.finalize()
        self.stop_time = time.time()
        Timeline.end("train")

    def record_details(self, json_name=None, meta=None):
        ## for the uber master, save yourself
        out_dict = { "train_time" : self.stop_time - self.start_time,
                     "history":self.histories}
        if meta is not None:
            out_dict.update( meta )

        if not json_name:
            json_name = 'master-history-%s-%s.json'%(os.getpid(), self.start_time)

        if self.model:
            model_file = json_name.replace('.json','.model')
            self.model.save( model_file )

        if self.algo:
            algo_file = json_name.replace('.json','.algo')
            self.algo.save( algo_file )

        with open( json_name, 'w') as out_file:
            out_file.write( json.dumps(out_dict, indent=4, separators=(',',': ')) )

        return self.histories

    def validation_worker(self):
        """Main function of the validation thread"""
        logging.debug("Validation thread started")
        while True:
            item = self.validation_queue.get()
            if item is None:
                logging.debug("Validation thread signing off")
                self.validation_queue.task_done()
                break
            weights, model = item
            try:
                self.validate_aux(weights, model)
            except Exception as e:
                logging.error(e)
            self.validation_queue.task_done()

    def validate(self, weights):
        if self.threaded_validation:
            model = self.validation_model
            self.validation_queue.put((weights, model))
        else:
            return self.validate_aux(weights, self.model)
        
    
    def validate_aux(self, weights, model):
        """Compute the loss on the validation data.
            Return a dictionary of validation metrics."""
        Timeline.begin("validation", "VALIDATION")
        if self.has_parent:
            return {}
        model.set_weights(weights)

        logging.debug("Starting validation")
        val_metrics = np.zeros((1,))
        i_batch = 0
        for i_batch, batch in enumerate(self.data.generate_data()):
            new_val_metrics =  model.test_on_batch(x=batch[0], y =batch[1] )
            if val_metrics.shape != new_val_metrics.shape:
                val_metrics =  np.zeros(new_val_metrics.shape)
            val_metrics += new_val_metrics
            if self._short_batches and i_batch>self._short_batches: break
        val_metrics = val_metrics / float(i_batch+1)
        l = model.get_logs(val_metrics, val=True)
        self.update_history( l )

        if self.target_metric:
            m,opp,v = self.target_metric
            model,m= m.split(':') if ':' in m else (None,m)
            r = self.history_key()
            use = self.histories[r].get(model,None) if model else self.histories[r]
            if use:
                if m in use and ((opp=='>' and use[m][-1]>v) or (opp=='<' and use[m][-1]<v)):
                    logging.debug("metric %s is %s %s, stop training",str(m),str(opp), str(v))
                    self.stop_training = True
            else:
                logging.error("fatal target stopping cannot get %s", str(m))
                MPI.COMM_WORLD.Abort()
                
        if self.patience:
            m,opp,p = self.patience
            p = int(p)
            model,m=m.split(':') if ':' in m else (None,m)
            r = self.history_key()            
            use = self.histories[r].get(model,None) if model else self.histories[r]
            ref = None
            smooth = False
            if use:
                if m in use:
                    if '~' in opp:
                        smooth = True
                        opp = opp.replace('~','')
                        ## do the averaging
                        if len(use[m])>=(2*p):
                            ref = np.mean(use[m][-(2*p):-p])
                            current = np.mean(use[m][-p:])
                        else:
                            if len(use[m])>=p:
                                ref = use[m][-p]
                                current = use[m][-1]
                                current = None
                    if ref is not None and current is not None and ((ref<current and opp=='<') or (ref>current and opp=='>')):
                        logging.info("metric %s is over %s patience boundary: %s (ref) %s %s (current) %s", str(m), str(p), str(ref), str(opp), str(current), "with smoothing" if smooth else "")
                        self.stop_training = True
                else:
                    logging.error("fatal early stopping cannot get %s", str(m))
                    MPI.COMM_WORLD.Abort()
            else:
                logging.error("fatal early stopping cannot get %s", str(m))
                MPI.COMM_WORLD.Abort()
        logging.info("Validation metrics:")
        self.print_metrics(val_metrics)
        logging.debug("Ending validation")
        Timeline.end("validation", "VALIDATION")
        return None

    ### MPI-related functions below

    def signal_children(self):
        """Sends each child a message telling them to start training"""
        for child in range(1, self.child_comm.Get_size()):
            self.send( obj=None, tag='train', dest=child, comm=self.child_comm )

    def recv_any_from_child(self,status):
        """Receives any message from any child.  Returns the provided status object,
            populated with information about received message"""
        self.recv( tag='any', source=MPI.ANY_SOURCE, status=status, comm=self.child_comm )
        return status

    def recv_history_from_child(self, child):
        return self.recv( tag='history', source=child, comm=self.child_comm )

    def send_exit_to_child(self, child, comm=None):
        if comm is None:
            comm = self.child_comm
        return comm.isend( None, dest=child, tag=self.lookup_mpi_tag('exit') )
    
    def build_model(self):
        """Builds the Keras model and updates model-related attributes"""

        if self.threaded_validation:
            self.validation_model = self.model_builder.build_model()
            self.algo.compile_model( self.validation_model )
        super(MPIMaster, self).build_model(local_session=self.threaded_validation)



### Monitor class

import os
import logging
from threading import Thread
import psutil
try:
    import pynvml
except:
    logging.warning("pynvml does not load, no monitoring available")
import time

class Monitor(object):
    """ Class that monitors GPU utilization for a given time """

    def __init__(self, sampling_rate=None, pid=None):
        self.sampling_rate = sampling_rate if sampling_rate is not None else 0.5
        self.pid = pid if pid is not None else os.getpid()
        
        self.gpu = None
        self.should_stop = False
        self.thread = None
        self.accounting_enabled = False
        self.stats = []
    
    def _find_gpu(self):
        device_count = pynvml.nvmlDeviceGetCount()
        for i in range(device_count):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            gpu_processes = pynvml.nvmlDeviceGetComputeRunningProcesses(handle)
            for gpu_process in gpu_processes:
                if gpu_process.pid == self.pid:
                    self.gpu = handle
        
        self.accounting_enabled = pynvml.nvmlDeviceGetAccountingMode(self.gpu) == pynvml.NVML_FEATURE_ENABLED
        
        # Clear accounting statistics (requires root privileges)
        #pynvml.nvmlDeviceSetAccountingMode(self.gpu, pynvml.NVML_FEATURE_DISABLED)
        #pynvml.nvmlDeviceSetAccountingMode(self.gpu, pynvml.NVML_FEATURE_ENABLED)

    def _monitor(self):
        pynvml.nvmlInit()
        self._find_gpu()
        current_sample = []
        while not self.should_stop:
            used_cpu = None
            used_cpumem = None
            used_gpu = None
            used_gpumem = None
            
            cpu_process = psutil.Process(self.pid)
            used_cpu = cpu_process.cpu_percent() / psutil.cpu_count() # CPU utilization in %
            used_cpumem = cpu_process.memory_info().rss // (1024*1024) # Memory use in MB

            gpu_processes = pynvml.nvmlDeviceGetComputeRunningProcesses(self.gpu)
            for gpu_process in gpu_processes:
                if gpu_process.pid == self.pid:
                    used_gpumem = gpu_process.usedGpuMemory // (1024*1024) # GPU memory use in MB
                    break

            if self.accounting_enabled:
                try:
                    stats = pynvml.nvmlDeviceGetAccountingStats(self.gpu, self.pid)
                    used_gpu = stats.gpuUtilization
                except pynvml.NVMLError: # NVMLError_NotFound
                    pass

            if not used_gpu:
                util = pynvml.nvmlDeviceGetUtilizationRates(self.gpu)
                used_gpu = util.gpu / len(gpu_processes) # Approximate based on number of processes

            current_sample.append((used_cpu, used_cpumem, used_gpu, used_gpumem))

            time.sleep(self.sampling_rate)

        self.stats.append([round(sum(x) / len(x)) for x in zip(*current_sample)])
        pynvml.nvmlShutdown()
    
    def start_monitor(self):
        self.should_stop = False
        self.thread = Thread(target=self._monitor)
        self.thread.start()
    
    def stop_monitor(self):
        self.should_stop = True
        self.thread.join()
    
    def get_stats(self):
        return self.stats

